/* MiniDLNA project
 *
 * http://sourceforge.net/projects/minidlna/
 *
 * MiniDLNA media server
 * Copyright (C) 2008-2009  Justin Maggard
 *
 * This file is part of MiniDLNA.
 *
 * MiniDLNA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * MiniDLNA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MiniDLNA. If not, see <http://www.gnu.org/licenses/>.
 *
 * Portions of the code from the MiniUPnP project:
 *
 * Copyright (c) 2006-2007, Thomas Bernard
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>

#include "upnpglobalvars.h"
#include "utils.h"
#include "upnphttp.h"
#include "upnpsoap.h"
#include "containers.h"
#include "upnpreplyparse.h"
#include "getifaddr.h"
#include "scanner.h"
#include "sql.h"
#include "log.h"


#ifdef ABILITY_MODIFICATION
#include <dlna_stringtable.h>
#include <hce_DLNAWiFiInterface.h>
//#include <Network_LinuxWiFiUISysInterface.h>
#endif
#ifdef __sparc__ /* Sorting takes too long on slow processors with very large containers */
# define __SORT_LIMIT if( totalMatches < 10000 )
#else
# define __SORT_LIMIT
#endif

/* Standard Errors:
 *
 * errorCode errorDescription Description
 * --------	---------------- -----------
 * 401 		Invalid Action 	No action by that name at this service.
 * 402 		Invalid Args 	Could be any of the following: not enough in args,
 * 							too many in args, no in arg by that name, 
 * 							one or more in args are of the wrong data type.
 * 403 		Out of Sync 	Out of synchronization.
 * 501 		Action Failed 	May be returned in current state of service
 * 							prevents invoking that action.
 * 600-699 	TBD 			Common action errors. Defined by UPnP Forum
 * 							Technical Committee.
 * 700-799 	TBD 			Action-specific errors for standard actions.
 * 							Defined by UPnP Forum working committee.
 * 800-899 	TBD 			Action-specific errors for non-standard actions. 
 * 							Defined by UPnP vendor.
*/
static void
SoapError(struct upnphttp * h, int errCode, const char * errDesc)
{
	static const char resp[] = 
		"<s:Envelope "
		"xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" "
		"s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
		"<s:Body>"
		"<s:Fault>"
		"<faultcode>s:Client</faultcode>"
		"<faultstring>UPnPError</faultstring>"
		"<detail>"
		"<UPnPError xmlns=\"urn:schemas-upnp-org:control-1-0\">"
		"<errorCode>%d</errorCode>"
		"<errorDescription>%s</errorDescription>"
		"</UPnPError>"
		"</detail>"
		"</s:Fault>"
		"</s:Body>"
		"</s:Envelope>";

	char body[2048];
	int bodylen;

	DPRINTF(E_WARN, L_HTTP, "Returning UPnPError %d: %s\n", errCode, errDesc);
	bodylen = snprintf(body, sizeof(body), resp, errCode, errDesc);
	BuildResp2_upnphttp(h, 500, "Internal Server Error", body, bodylen);
	SendResp_upnphttp(h);
	CloseSocket_upnphttp(h);
}

static void
BuildSendAndCloseSoapResp(struct upnphttp * h,
                          const char * body, int bodylen)
{
	static const char beforebody[] =
		"<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
		"<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" "
		"s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
		"<s:Body>";

	static const char afterbody[] =
		"</s:Body>"
		"</s:Envelope>\r\n";

	if (!body || bodylen < 0)
	{
		Send500(h);
		return;
	}

	BuildHeader_upnphttp(h, 200, "OK",  sizeof(beforebody) - 1
		+ sizeof(afterbody) - 1 + bodylen );

	memcpy(h->res_buf + h->res_buflen, beforebody, sizeof(beforebody) - 1);
	h->res_buflen += sizeof(beforebody) - 1;

	memcpy(h->res_buf + h->res_buflen, body, bodylen);
	h->res_buflen += bodylen;

	memcpy(h->res_buf + h->res_buflen, afterbody, sizeof(afterbody) - 1);
	h->res_buflen += sizeof(afterbody) - 1;

	SendResp_upnphttp(h);
	CloseSocket_upnphttp(h);
}

static void
GetSystemUpdateID(struct upnphttp * h, const char * action)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<Id>%d</Id>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp,
		action, "urn:schemas-upnp-org:service:ContentDirectory:1",
		updateID, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
IsAuthorizedValidated(struct upnphttp * h, const char * action)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<Result>%d</Result>"
		"</u:%sResponse>";

	char body[512];
	struct NameValueParserData data;
	const char * id;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data, XML_STORE_EMPTY_FL);
	id = GetValueFromNameValueList(&data, "DeviceID");
	if(id)
	{
		int bodylen;
		bodylen = snprintf(body, sizeof(body), resp,
			action, "urn:microsoft.com:service:X_MS_MediaReceiverRegistrar:1",
			1, action);	
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}
	else
		SoapError(h, 402, "Invalid Args");

	ClearNameValueList(&data);	
}

static void
RegisterDevice(struct upnphttp * h, const char * action)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<RegistrationRespMsg>%s</RegistrationRespMsg>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp,
		action, "urn:microsoft.com:service:X_MS_MediaReceiverRegistrar:1",
		uuidvalue, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetProtocolInfo(struct upnphttp * h, const char * action)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<Source>"
		RESOURCE_PROTOCOL_INFO_VALUES
		"</Source>"
		"<Sink></Sink>"
		"</u:%sResponse>";

	char * body;
	int bodylen;

	bodylen = asprintf(&body, resp,
		action, "urn:schemas-upnp-org:service:ConnectionManager:1",
		action);	
	BuildSendAndCloseSoapResp(h, body, bodylen);
	free(body);
}

static void
GetSortCapabilities(struct upnphttp * h, const char * action)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<SortCaps>"
		  "dc:title,"
		  "dc:date,"
		  "upnp:class,"
		  "upnp:album,"
		  "upnp:originalTrackNumber"
		"</SortCaps>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp,
		action, "urn:schemas-upnp-org:service:ContentDirectory:1",
		action);	
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetSearchCapabilities(struct upnphttp * h, const char * action)
{
	static const char resp[] =
		"<u:%sResponse xmlns:u=\"%s\">"
		"<SearchCaps>"
		  "dc:creator,"
		  "dc:date,"
		  "dc:title,"
		  "upnp:album,"
		  "upnp:actor,"
		  "upnp:artist,"
		  "upnp:class,"
		  "upnp:genre,"
		  "@id,"
		  "@parentID,"
		  "@refID"
		"</SearchCaps>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp,
		action, "urn:schemas-upnp-org:service:ContentDirectory:1",
		action);	
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetCurrentConnectionIDs(struct upnphttp * h, const char * action)
{
	/* TODO: Use real data. - JM */
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<ConnectionIDs>0</ConnectionIDs>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp,
		action, "urn:schemas-upnp-org:service:ConnectionManager:1",
		action);	
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetCurrentConnectionInfo(struct upnphttp * h, const char * action)
{
	/* TODO: Use real data. - JM */
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<RcsID>-1</RcsID>"
		"<AVTransportID>-1</AVTransportID>"
		"<ProtocolInfo></ProtocolInfo>"
		"<PeerConnectionManager></PeerConnectionManager>"
		"<PeerConnectionID>-1</PeerConnectionID>"
		"<Direction>Output</Direction>"
		"<Status>Unknown</Status>"
		"</u:%sResponse>";

	char body[sizeof(resp)+128];
	struct NameValueParserData data;
	const char *id_str;
	int id;
	char *endptr = NULL;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data, XML_STORE_EMPTY_FL);
	id_str = GetValueFromNameValueList(&data, "ConnectionID");
	DPRINTF(E_INFO, L_HTTP, "GetCurrentConnectionInfo(%s)\n", id_str);
	if(id_str)
		id = strtol(id_str, &endptr, 10);
	if (!id_str || endptr == id_str)
	{
		SoapError(h, 402, "Invalid Args");
	}
	else if(id != 0)
	{
		SoapError(h, 701, "No such object error");
	}
	else
	{
		int bodylen;
		bodylen = snprintf(body, sizeof(body), resp,
			action, "urn:schemas-upnp-org:service:ConnectionManager:1",
			action);	
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}
	ClearNameValueList(&data);	
}

/* Standard DLNA/UPnP filter flags */
#define FILTER_CHILDCOUNT                        0x00000001
#define FILTER_DC_CREATOR                        0x00000002
#define FILTER_DC_DATE                           0x00000004
#define FILTER_DC_DESCRIPTION                    0x00000008
#define FILTER_DLNA_NAMESPACE                    0x00000010
#define FILTER_REFID                             0x00000020
#define FILTER_RES                               0x00000040
#define FILTER_RES_BITRATE                       0x00000080
#define FILTER_RES_DURATION                      0x00000100
#define FILTER_RES_NRAUDIOCHANNELS               0x00000200
#define FILTER_RES_RESOLUTION                    0x00000400
#define FILTER_RES_SAMPLEFREQUENCY               0x00000800
#define FILTER_RES_SIZE                          0x00001000
#define FILTER_SEARCHABLE                        0x00002000
#define FILTER_UPNP_ACTOR                        0x00004000
#define FILTER_UPNP_ALBUM                        0x00008000
#define FILTER_UPNP_ALBUMARTURI                  0x00010000
#define FILTER_UPNP_ALBUMARTURI_DLNA_PROFILEID   0x00020000
#define FILTER_UPNP_ARTIST                       0x00040000
#define FILTER_UPNP_GENRE                        0x00080000
#define FILTER_UPNP_ORIGINALTRACKNUMBER          0x00100000
#define FILTER_UPNP_SEARCHCLASS                  0x00200000
#define FILTER_UPNP_STORAGEUSED                  0x00400000
/* Vendor-specific filter flags */
#define FILTER_SEC_CAPTION_INFO_EX               0x01000000
#define FILTER_SEC_DCM_INFO                      0x02000000
#define FILTER_PV_SUBTITLE_FILE_TYPE             0x04000000
#define FILTER_PV_SUBTITLE_FILE_URI              0x08000000
#define FILTER_PV_SUBTITLE                       0x0C000000
#define FILTER_AV_MEDIA_CLASS                    0x10000000

static uint32_t
set_filter_flags(char *filter, struct upnphttp *h)
{
	char *item, *saveptr = NULL;
	uint32_t flags = 0;
	int samsung = h->req_client && (h->req_client->type->flags & FLAG_SAMSUNG);

	if( !filter || (strlen(filter) <= 1) ) {
		/* Not the full 32 bits.  Skip vendor-specific stuff by default. */
		flags = 0xFFFFFF;
		if (samsung)
			flags |= FILTER_SEC_CAPTION_INFO_EX | FILTER_SEC_DCM_INFO;
	}
	if (flags)
		return flags;

	if( samsung )
		flags |= FILTER_DLNA_NAMESPACE;
	item = strtok_r(filter, ",", &saveptr);
	while( item != NULL )
	{
		if( saveptr )
			*(item-1) = ',';
		while( isspace(*item) )
			item++;
		if( strcmp(item, "@childCount") == 0 )
		{
			flags |= FILTER_CHILDCOUNT;
		}
		else if( strcmp(item, "@searchable") == 0 )
		{
			flags |= FILTER_SEARCHABLE;
		}
		else if( strcmp(item, "dc:creator") == 0 )
		{
			flags |= FILTER_DC_CREATOR;
		}
		else if( strcmp(item, "dc:date") == 0 )
		{
			flags |= FILTER_DC_DATE;
		}
		else if( strcmp(item, "dc:description") == 0 )
		{
			flags |= FILTER_DC_DESCRIPTION;
		}
		else if( strcmp(item, "dlna") == 0 )
		{
			flags |= FILTER_DLNA_NAMESPACE;
		}
		else if( strcmp(item, "@refID") == 0 )
		{
			flags |= FILTER_REFID;
		}
		else if( strcmp(item, "upnp:album") == 0 )
		{
			flags |= FILTER_UPNP_ALBUM;
		}
		else if( strcmp(item, "upnp:albumArtURI") == 0 )
		{
			flags |= FILTER_UPNP_ALBUMARTURI;
			if( samsung )
				flags |= FILTER_UPNP_ALBUMARTURI_DLNA_PROFILEID;
		}
		else if( strcmp(item, "upnp:albumArtURI@dlna:profileID") == 0 )
		{
			flags |= FILTER_UPNP_ALBUMARTURI;
			flags |= FILTER_UPNP_ALBUMARTURI_DLNA_PROFILEID;
		}
		else if( strcmp(item, "upnp:artist") == 0 )
		{
			flags |= FILTER_UPNP_ARTIST;
		}
		else if( strcmp(item, "upnp:actor") == 0 )
		{
			flags |= FILTER_UPNP_ACTOR;
		}
		else if( strcmp(item, "upnp:genre") == 0 )
		{
			flags |= FILTER_UPNP_GENRE;
		}
		else if( strcmp(item, "upnp:originalTrackNumber") == 0 )
		{
			flags |= FILTER_UPNP_ORIGINALTRACKNUMBER;
		}
		else if( strcmp(item, "upnp:searchClass") == 0 )
		{
			flags |= FILTER_UPNP_SEARCHCLASS;
		}
		else if( strcmp(item, "upnp:storageUsed") == 0 )
		{
			flags |= FILTER_UPNP_STORAGEUSED;
		}
		else if( strcmp(item, "res") == 0 )
		{
			flags |= FILTER_RES;
		}
		else if( (strcmp(item, "res@bitrate") == 0) ||
		         (strcmp(item, "@bitrate") == 0) ||
		         ((strcmp(item, "bitrate") == 0) && (flags & FILTER_RES)) )
		{
			flags |= FILTER_RES;
			flags |= FILTER_RES_BITRATE;
		}
		else if( (strcmp(item, "res@duration") == 0) ||
		         (strcmp(item, "@duration") == 0) ||
		         ((strcmp(item, "duration") == 0) && (flags & FILTER_RES)) )
		{
			flags |= FILTER_RES;
			flags |= FILTER_RES_DURATION;
		}
		else if( (strcmp(item, "res@nrAudioChannels") == 0) ||
		         (strcmp(item, "@nrAudioChannels") == 0) ||
		         ((strcmp(item, "nrAudioChannels") == 0) && (flags & FILTER_RES)) )
		{
			flags |= FILTER_RES;
			flags |= FILTER_RES_NRAUDIOCHANNELS;
		}
		else if( (strcmp(item, "res@resolution") == 0) ||
		         (strcmp(item, "@resolution") == 0) ||
		         ((strcmp(item, "resolution") == 0) && (flags & FILTER_RES)) )
		{
			flags |= FILTER_RES;
			flags |= FILTER_RES_RESOLUTION;
		}
		else if( (strcmp(item, "res@sampleFrequency") == 0) ||
		         (strcmp(item, "@sampleFrequency") == 0) ||
		         ((strcmp(item, "sampleFrequency") == 0) && (flags & FILTER_RES)) )
		{
			flags |= FILTER_RES;
			flags |= FILTER_RES_SAMPLEFREQUENCY;
		}
		else if( (strcmp(item, "res@size") == 0) ||
		         (strcmp(item, "@size") == 0) ||
		         (strcmp(item, "size") == 0) )
		{
			flags |= FILTER_RES;
			flags |= FILTER_RES_SIZE;
		}
		else if( strcmp(item, "sec:CaptionInfoEx") == 0 )
		{
			flags |= FILTER_SEC_CAPTION_INFO_EX;
		}
		else if( strcmp(item, "sec:dcmInfo") == 0 )
		{
			flags |= FILTER_SEC_DCM_INFO;
		}
		else if( strcmp(item, "res@pv:subtitleFileType") == 0 )
		{
			flags |= FILTER_PV_SUBTITLE_FILE_TYPE;
		}
		else if( strcmp(item, "res@pv:subtitleFileUri") == 0 )
		{
			flags |= FILTER_PV_SUBTITLE_FILE_URI;
		}
		else if( strcmp(item, "av:mediaClass") == 0 )
		{
			flags |= FILTER_AV_MEDIA_CLASS;
		}
		item = strtok_r(NULL, ",", &saveptr);
	}

	return flags;
}

static char *
parse_sort_criteria(char *sortCriteria, int *error)
{
	char *order = NULL;
	char *item, *saveptr;
	int i, ret, reverse, title_sorted = 0;
	struct string_s str;
	*error = 0;

	if( force_sort_criteria )
		sortCriteria = strdup(force_sort_criteria);
	if( !sortCriteria )
		return NULL;

	if( (item = strtok_r(sortCriteria, ",", &saveptr)) )
	{
		order = malloc(4096);
		str.data = order;
		str.size = 4096;
		str.off = 0;
		strcatf(&str, "order by ");
	}
	for( i = 0; item != NULL; i++ )
	{
		reverse=0;
		if( i )
			strcatf(&str, ", ");
		if( *item == '+' )
		{
			item++;
		}
		else if( *item == '-' )
		{
			reverse = 1;
			item++;
		}
		else
		{
			DPRINTF(E_ERROR, L_HTTP, "No order specified [%s]\n", item);
			goto bad_direction;
		}
		if( strcasecmp(item, "upnp:class") == 0 )
		{
			strcatf(&str, "o.CLASS");
		}
		else if( strcasecmp(item, "dc:title") == 0 )
		{
			strcatf(&str, "d.TITLE");
			title_sorted = 1;
		}
		else if( strcasecmp(item, "dc:date") == 0 )
		{
			strcatf(&str, "d.DATE");
		}
		else if( strcasecmp(item, "upnp:originalTrackNumber") == 0 )
		{
			strcatf(&str, "d.DISC, d.TRACK");
		}
		else if( strcasecmp(item, "upnp:album") == 0 )
		{
			strcatf(&str, "d.ALBUM");
		}
		else
		{
			DPRINTF(E_ERROR, L_HTTP, "Unhandled SortCriteria [%s]\n", item);
		bad_direction:
			*error = -1;
			if( i )
			{
				ret = strlen(order);
				order[ret-2] = '\0';
			}
			i--;
			goto unhandled_order;
		}

		if( reverse )
			strcatf(&str, " DESC");
		unhandled_order:
		item = strtok_r(NULL, ",", &saveptr);
	}
	if( i <= 0 )
	{
		free(order);
		if( force_sort_criteria )
			free(sortCriteria);
		return NULL;
	}
	/* Add a "tiebreaker" sort order */
	if( !title_sorted )
		strcatf(&str, ", TITLE ASC");

	if( force_sort_criteria )
		free(sortCriteria);

	return order;
}

inline static void
add_resized_res(int srcw, int srch, int reqw, int reqh, char *dlna_pn,
                char *detailID, struct Response *args)
{
	int dstw = reqw;
	int dsth = reqh;

	if( (args->flags & FLAG_NO_RESIZE) && reqw > 160 && reqh > 160 )
		return;

	strcatf(args->str, "&lt;res ");
	if( args->filter & FILTER_RES_RESOLUTION )
	{
		dstw = reqw;
		dsth = ((((reqw<<10)/srcw)*srch)>>10);
		if( dsth > reqh ) {
			dsth = reqh;
			dstw = (((reqh<<10)/srch) * srcw>>10);
		}
		strcatf(args->str, "resolution=\"%dx%d\" ", dstw, dsth);
	}
    
    #ifdef ABILITY_MODIFICATION
    //Swtich the content type to be size index
    EnumCustomizeTableIndex eSizeStringIndex = PANA_TABLE_INDEX_BROWSE_SIZE_ORG;
    //4.!!@@## need to modify the value of eSizeStringIndex according to file size.
	DPRINTF(E_DEBUG, L_GENERAL, "[PPD][%s %d]size :%s \n",__FUNCTION__,__LINE__,dlna_pn);
	strcatf(args->str, "protocolInfo=\"http-get:*:image/jpeg:"
	                          "DLNA.ORG_PN=%s;"
	                          "DLNA.ORG_CI=1;DLNA.ORG_FLAGS=%08X%024X;"
	                          "%s=%s;"
	                          "\"&gt;http://%s:%d"
	                          "/Resized/%s.jpg?width=%d,height=%d"
	                          "&lt;/res&gt;",
	                          dlna_pn, 
	                          DLNA_FLAG_DLNA_V1_5|DLNA_FLAG_HTTP_STALLING|DLNA_FLAG_TM_B|DLNA_FLAG_TM_I,
	                          0,
	                          GetDLNACustomizedTableString(PANA_TABLE_INDEX_BROWSE_TAG_PANA_SIZE),
	                          GetDLNACustomizedTableString(eSizeStringIndex),
	                          lan_addr[args->iface].str,runtime_vars.port,
	                          detailID, 
	                          dstw, 
	                          dsth);
    #else
	strcatf(args->str, "protocolInfo=\"http-get:*:image/jpeg:"
	                          "DLNA.ORG_PN=%s;DLNA.ORG_CI=1;DLNA.ORG_FLAGS=%08X%024X\"&gt;"
	                          "http://%s:%d/Resized/%s.jpg?width=%d,height=%d"
	                          "&lt;/res&gt;",
	                          dlna_pn, DLNA_FLAG_DLNA_V1_5|DLNA_FLAG_HTTP_STALLING|DLNA_FLAG_TM_B|DLNA_FLAG_TM_I, 0,
	                          lan_addr[args->iface].str, runtime_vars.port,
	                          detailID, dstw, dsth);
    #endif

    
}

inline static void
add_res(char *size, char *duration, char *bitrate, char *sampleFrequency,
        char *nrAudioChannels, char *resolution, char *dlna_pn, char *mime,
        char *detailID, const char *ext, struct Response *args)
{
	strcatf(args->str, "&lt;res ");
	if( size && (args->filter & FILTER_RES_SIZE) ) {
		strcatf(args->str, "size=\"%s\" ", size);
	}
	if( duration && (args->filter & FILTER_RES_DURATION) ) {
		strcatf(args->str, "duration=\"%s\" ", duration);
	}
	if( bitrate && (args->filter & FILTER_RES_BITRATE) ) {
		int br = atoi(bitrate);
		if(args->flags & FLAG_MS_PFS)
			br /= 8;
		strcatf(args->str, "bitrate=\"%d\" ", br);
	}
	if( sampleFrequency && (args->filter & FILTER_RES_SAMPLEFREQUENCY) ) {
		strcatf(args->str, "sampleFrequency=\"%s\" ", sampleFrequency);
	}
	if( nrAudioChannels && (args->filter & FILTER_RES_NRAUDIOCHANNELS) ) {
		strcatf(args->str, "nrAudioChannels=\"%s\" ", nrAudioChannels);
	}
	if( resolution && (args->filter & FILTER_RES_RESOLUTION) ) {
		strcatf(args->str, "resolution=\"%s\" ", resolution);
	}
	if( args->filter & FILTER_PV_SUBTITLE )
	{
		if( args->flags & FLAG_HAS_CAPTIONS )
		{
			if( args->filter & FILTER_PV_SUBTITLE_FILE_TYPE )
				strcatf(args->str, "pv:subtitleFileType=\"SRT\" ");
			if( args->filter & FILTER_PV_SUBTITLE_FILE_URI )
				strcatf(args->str, "pv:subtitleFileUri=\"http://%s:%d/Captions/%s.srt\" ",
			                lan_addr[args->iface].str, runtime_vars.port, detailID);
		}
	}
    
	DPRINTF(E_DEBUG, L_GENERAL, "[PPD][%s %d] \n",__FUNCTION__,__LINE__);
	strcatf(args->str, "protocolInfo=\"http-get:*:%s:%s\"&gt;"
	                          "http://%s:%d/MediaItems/%s.%s"
	                          "&lt;/res&gt;",
	                          mime, dlna_pn, lan_addr[args->iface].str,
	                          runtime_vars.port, detailID, ext);
}

static int
get_child_count(const char *object, struct magic_container_s *magic)
{
	int ret;

	if (magic && magic->child_count)
		ret = sql_get_int_field(db, "SELECT count(*) from %s", magic->child_count);
	else if (magic && magic->objectid && *(magic->objectid))
		ret = sql_get_int_field(db, "SELECT count(*) from OBJECTS where PARENT_ID = '%s';", *(magic->objectid));
	else
		ret = sql_get_int_field(db, "SELECT count(*) from OBJECTS where PARENT_ID = '%s';", object);

	return (ret > 0) ? ret : 0;
}

static int
object_exists(const char *object)
{
	int ret;
	ret = sql_get_int_field(db, "SELECT count(*) from OBJECTS where OBJECT_ID = '%q'",
				strcmp(object, "*") == 0 ? "0" : object);
	return (ret > 0);
}

#define COLUMNS "o.DETAIL_ID, o.CLASS," \
                " d.SIZE, d.TITLE, d.DURATION, d.BITRATE, d.SAMPLERATE, d.ARTIST," \
                " d.ALBUM, d.GENRE, d.COMMENT, d.CHANNELS, d.TRACK, d.DATE, d.RESOLUTION," \
                " d.THUMBNAIL, d.CREATOR, d.DLNA_PN, d.MIME, d.ALBUM_ART, d.ROTATION, d.DISC "
#define SELECT_COLUMNS "SELECT o.OBJECT_ID, o.PARENT_ID, o.REF_ID, " COLUMNS

#define NON_ZERO(x) (x && atoi(x))
#define IS_ZERO(x) (!x || !atoi(x))
#ifdef ABILITY_MODIFICATION
typedef struct _UpnpSoapBrowseItemStructure_  {
    char                        pbFileType[16];
    char*                       pbFileFullPath;
    char                        pbResolution[16];
    DLNAWiFiFileSizeType        eFileSize;
    DLNAWiFiVideoFormat         eVideoFormat;
    DLNAWiFiVideoType           eVideoType;
    unsigned int                ulFileSize;
    uint8_t                     bIsResized;
} UpnpSoapBrowseItemStructure;
#define NAMINGLENGTH (11)   //ex: 1000001.JPG
int GetBrowseName(DLNAWiFiFileSizeType FileType, char* FileName, char* BrowseName)
{
    switch(FileType)
    {
        case DLNA_WIFI_FILE_SIZE_TYPE_ORG:
        {
            memcpy(BrowseName, "/DO", strlen("/DO"));
            strncat(BrowseName, FileName+strlen(FileName)-NAMINGLENGTH, NAMINGLENGTH);
            DPRINTF(E_DEBUG, L_GENERAL, "%s\n", BrowseName);
        }
        break;

        case DLNA_WIFI_FILE_SIZE_TYPE_TN:
        {
            memcpy(BrowseName, "/DT", strlen("/DT"));
            strncat(BrowseName, FileName+strlen(FileName)-NAMINGLENGTH, NAMINGLENGTH);
            DPRINTF(E_DEBUG, L_GENERAL, "%s\n", BrowseName);
        }
        break;

        case DLNA_WIFI_FILE_SIZE_TYPE_LRG:
        {
            memcpy(BrowseName, "/DL", strlen("/DL"));
            strncat(BrowseName, FileName+strlen(FileName)-NAMINGLENGTH, NAMINGLENGTH);
            DPRINTF(E_DEBUG, L_GENERAL, "%s\n", BrowseName);
        }
        break;

        default:
        {
            DPRINTF(E_ERROR, L_GENERAL, "Invalid File Type!\n");
            return -1;
        }
        break;
    }
    return 0;
}

inline static int
AbilityDLNAAddResizedRes(UpnpSoapBrowseItemStructure* psItemInfo, struct Response *args)
{

	strcatf(args->str, "&lt;res ");
    //Swtich the content type to be size index
    EnumCustomizeTableIndex eSizeStringIndex = PANA_TABLE_INDEX_BROWSE_SIZE_ORG;
    char pbResizedFullPath[64] = {0};
    //char* pbStrLocat = NULL;
    int   bIsResized = psItemInfo->bIsResized;
    switch(psItemInfo->eFileSize)
    {
        case DLNA_WIFI_FILE_SIZE_TYPE_ORG:
            {
                eSizeStringIndex = PANA_TABLE_INDEX_BROWSE_SIZE_ORG;
                GetBrowseName(psItemInfo->eFileSize, psItemInfo->pbFileFullPath, pbResizedFullPath);
            }  
            break;
        case DLNA_WIFI_FILE_SIZE_TYPE_TN:
            {
                eSizeStringIndex = PANA_TABLE_INDEX_BROWSE_SIZE_TN;
                GetBrowseName(psItemInfo->eFileSize, psItemInfo->pbFileFullPath, pbResizedFullPath);
            }    
            break;
        case DLNA_WIFI_FILE_SIZE_TYPE_LRG:
            {
                eSizeStringIndex = PANA_TABLE_INDEX_BROWSE_SIZE_LRG;
                GetBrowseName(psItemInfo->eFileSize, psItemInfo->pbFileFullPath, pbResizedFullPath);
            }
            break;
        default :
	        DLNA_Printf(E_WARN, L_GENERAL, "[PPD] WARNING! not hand size type:%d\n",psItemInfo->eFileSize);
            break;
    }    
    //4.!!@@## Need Video type ex:CAM_AVC_MP4_HP_1080_30P_AAC
	DLNA_Printf(E_DEBUG, L_GENERAL, "[PPD]CP \n");
    
    if(strlen(psItemInfo->pbFileType) != 0)  {
    	strcatf(args->str, "protocolInfo=\"http-get:*:image/jpeg:"
                                  "DLNA.ORG_PN=%s;"
                                  "DLNA.ORG_OP=01;"
    	                          "DLNA.ORG_CI=%d;"
    	                          "DLNA.ORG_FLAGS=%08X%024X;"
    	                          "%s=%s\""
    	                          "&gt;http://%s:%d"
    	                          "%s"
    	                          "&lt;/res&gt;",
    	                          psItemInfo->pbFileType,
    	                          bIsResized,
    	                          //DLNA_FLAG_DLNA_V1_5|DLNA_FLAG_HTTP_STALLING|DLNA_FLAG_TM_B|DLNA_FLAG_TM_I,
    	                          DLNA_FLAG_DLNA_V1_5|DLNA_FLAG_TM_I,
    	                          0,
    	                          GetDLNACustomizedTableString(PANA_TABLE_INDEX_BROWSE_TAG_PANA_SIZE),
    	                          GetDLNACustomizedTableString(eSizeStringIndex),
    	                          lan_addr[args->iface].str,runtime_vars.port,
    	                          pbResizedFullPath
    	                          );
    }
    else  {
        strcatf(args->str, "protocolInfo=\"http-get:*:image/jpeg:"
                                  "DLNA.ORG_OP=01;"
    	                          "DLNA.ORG_FLAGS=%08X%024X;"
    	                          "%s=%s\""
    	                          "&gt;http://%s:%d"
    	                          "%s"
    	                          "&lt;/res&gt;",
    	                          //DLNA_FLAG_DLNA_V1_5|DLNA_FLAG_HTTP_STALLING|DLNA_FLAG_TM_B|DLNA_FLAG_TM_I,
    	                          DLNA_FLAG_DLNA_V1_5|DLNA_FLAG_TM_I,
    	                          0,
    	                          GetDLNACustomizedTableString(PANA_TABLE_INDEX_BROWSE_TAG_PANA_SIZE),
    	                          GetDLNACustomizedTableString(eSizeStringIndex),
    	                          lan_addr[args->iface].str,runtime_vars.port,
    	                          pbResizedFullPath
    	                          );
    }
    return 0;
}

inline static int
AbilityDLNAAddVideoRes(UpnpSoapBrowseItemStructure* psItemInfo, struct Response *args)
{

    strcatf(args->str, "&lt;res ");
    //Swtich the content type to be size index
    char PanaSizeString[33] = {0};
    char pbResizedFullPath[64] = {0};
    char* pbStrLocat = NULL;
    int   bIsResized = 0;

    if(psItemInfo->eVideoFormat == DLNA_WIFI_VIDEO_TYPE_MP4)
    {
        switch(psItemInfo->eVideoType)
        {
            case DLNA_WIFI_VIDEO_TYPE_1080P_60:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_1080_60P_AAC", strlen("CAM_AVC_MP4_MP_1080_60P_AAC"));
                }  
                break;
            case DLNA_WIFI_VIDEO_TYPE_1080P_50:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_1080_50P_AAC", strlen("CAM_AVC_MP4_MP_1080_50P_AAC"));
                }    
                break;
            case DLNA_WIFI_VIDEO_TYPE_1080P_30:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_1080_30P_AAC", strlen("CAM_AVC_MP4_MP_1080_30P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
            case DLNA_WIFI_VIDEO_TYPE_1080P_25:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_1080_25P_AAC", strlen("CAM_AVC_MP4_MP_1080_25P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_720P_60:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_720_60P_AAC", strlen("CAM_AVC_MP4_MP_720_60P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_720P_50:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_720_50P_AAC", strlen("CAM_AVC_MP4_MP_720_50P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_720P_30:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_720_30P_AAC", strlen("CAM_AVC_MP4_MP_720_30P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_720P_25:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_720_25P_AAC", strlen("CAM_AVC_MP4_MP_720_25P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_480P_30:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_848x480_30P_AAC", strlen("CAM_AVC_MP4_MP_848x480_30P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_480P_25:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_848x480_25P_AAC", strlen("CAM_AVC_MP4_MP_848x480_25P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
             case DLNA_WIFI_VIDEO_TYPE_360P_30:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_640x360_30P_AAC", strlen("CAM_AVC_MP4_MP_640x360_30P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
             case DLNA_WIFI_VIDEO_TYPE_360P_25:
                {
                    memcpy(PanaSizeString, "CAM_AVC_MP4_MP_640x360_25P_AAC", strlen("CAM_AVC_MP4_MP_640x360_25P_AAC"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            default :
                DLNA_Printf(E_WARN, L_GENERAL, "[PPD] WARNING! not hand size type:%d\n",psItemInfo->eFileSize);
                break;
        }        
    }
    else if(psItemInfo->eVideoFormat == DLNA_WIFI_VIDEO_TYPE_TS)
    {
        switch(psItemInfo->eVideoType)
        {
            case DLNA_WIFI_VIDEO_TYPE_1080P_60:
                {
                    memcpy(PanaSizeString, "CAM_AVC_TS_HP_1080_60P_AC3", strlen("CAM_AVC_TS_HP_1080_60P_AC3"));
                }  
                break;
            case DLNA_WIFI_VIDEO_TYPE_1080P_50:
                {
                    memcpy(PanaSizeString, "CAM_AVC_TS_HP_1080_50P_AC3", strlen("CAM_AVC_TS_HP_1080_50P_AC3"));
                }    
                break;
            case DLNA_WIFI_VIDEO_TYPE_1080P_30:
                {
                    memcpy(PanaSizeString, "CAM_AVC_TS_HP_1080_30P_AC3", strlen("CAM_AVC_TS_HP_1080_30P_AC3"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
            case DLNA_WIFI_VIDEO_TYPE_1080P_25:
                {
                    memcpy(PanaSizeString, "CAM_AVC_TS_HP_1080_25P_AC3", strlen("CAM_AVC_TS_HP_1080_25P_AC3"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_1080P_24:
                {
                    memcpy(PanaSizeString, "CAM_AVC_TS_HP_1080_24P_AC3", strlen("CAM_AVC_TS_HP_1080_24P_AC3"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_720P_60:
                {
                    memcpy(PanaSizeString, "CAM_AVC_TS_HP_720_60P_AC3", strlen("CAM_AVC_TS_HP_720_60P_AC3"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_720P_50:
                {
                    memcpy(PanaSizeString, "CAM_AVC_TS_HP_720_50P_AC3", strlen("CAM_AVC_TS_HP_720_50P_AC3"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_720P_30:
                {
                    memcpy(PanaSizeString, "CAM_AVC_TS_HP_720_30P_AC3", strlen("CAM_AVC_TS_HP_720_30P_AC3"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            case DLNA_WIFI_VIDEO_TYPE_720P_25:
                {
                    memcpy(PanaSizeString, "CAM_AVC_TS_HP_720_25P_AC3", strlen("CAM_AVC_TS_HP_720_25P_AC3"));
                    GetBrowseName(DLNA_WIFI_FILE_SIZE_TYPE_ORG, psItemInfo->pbFileFullPath, pbResizedFullPath);
                }
                break;
            default :
                DLNA_Printf(E_WARN, L_GENERAL, "[PPD] WARNING! not hand size type:%d\n",psItemInfo->eFileSize);
                break;
        }
    }

    if(psItemInfo->eFileSize == (DLNA_WIFI_VIDEO_TYPE_1080P_60 || DLNA_WIFI_VIDEO_TYPE_1080P_50))  {
        strcatf(args->str, "protocolInfo=\"http-get:*:application/octet-stream:"
                                  "DLNA.ORG_OP=01;"
                                  "%s=%s\""
                                  "&gt;http://%s:%d"
                                  "%s"
                                  "&lt;/res&gt;",
                                  0,
                                  GetDLNACustomizedTableString(PANA_TABLE_INDEX_BROWSE_TAG_PANA_SIZE),
                                  PanaSizeString,
                                  lan_addr[args->iface].str,runtime_vars.port,
                                  pbResizedFullPath
                                  );
        }
        else  {
            strcatf(args->str, "protocolInfo=\"http-get:*:video/mp4:"
                                      "DLNA.ORG_PN=%s;"
                                      "DLNA.ORG_OP=01;"
                                      "DLNA.ORG_CI=0;"
                                      "%s=%s\""
                                      "&gt;http://%s:%d"
                                      "%s"
                                      "&lt;/res&gt;",
                                      psItemInfo->pbFileType,
                                      0,
                                      GetDLNACustomizedTableString(PANA_TABLE_INDEX_BROWSE_TAG_PANA_SIZE),
                                      PanaSizeString,
                                      lan_addr[args->iface].str,runtime_vars.port,
                                      pbResizedFullPath
                                      );
        }
    return 0;
}


typedef struct _UpnpSoapBrowseContainerStructure_  {
    char                        pbID[16];
    char                        pbParent[16];
    char                        pbTitle[32];
    char                        pbClass[16];
    char                        pbFilePath[16];
    DLNAWiFiBrowseContainerType eContainerType;
    int                         lChildCount;
    HceWiFiDatabaseFileType     eFileType;
} UpnpSoapBrowseContainerStructure;   
static int UpnpSoapConstructContainerData(void *args, UpnpSoapBrowseContainerStructure *psInputInfo)
{
  int ret = 0;
  struct Response *passed_args = (struct Response *)args;
  struct string_s *str         = passed_args->str;
  
  DLNA_Printf(E_DEBUG, L_GENERAL,"[PPD] id : %s  parent : %s title:%s\n",psInputInfo->pbID, psInputInfo->pbParent,psInputInfo->pbTitle);
  DLNA_Printf(E_DEBUG, L_GENERAL,"[PPD] class : %s  path : %s CT:%d C:%d FT:%d\n",psInputInfo->pbClass, psInputInfo->pbFilePath, 
    psInputInfo->eContainerType, psInputInfo->lChildCount, psInputInfo->eFileType);
  
  //ret = strcatf(str, "&lt;container id=\"%s\" parentID=\"%s\" restricted=\"1\" ", psInputInfo->pbID, psInputInfo->pbParent);
  ret = strcatf(str, "&lt;container id=\"%s\" parentID=\"%s\" restricted=\"0\" childCount=\"%d\"&gt;", psInputInfo->pbID,
                psInputInfo->pbParent, psInputInfo->lChildCount);
  
  //passed_args->filter & FILTER_SEARCHABLE
  //ret = strcatf(str, " searchable=\"%d\"", 1);  
  //passed_args->filter & FILTER_CHILDCOUNT 
  //ret = strcatf(str, " childCount=\"%d\"", psInputInfo->lChildCount);
  
  /* If the client calls for BrowseMetadata on root, we have to include our "upnp:searchClass"'s, unless they're filtered out */
  if(psInputInfo->eContainerType == DLNA_WIFI_BROWSE_CONTAINER_TYPE_METADATA) {
  	//ret = strcatf(str, "&lt;upnp:searchClass includeDerived=\"1\"&gt;object.item.imageItem&lt;/upnp:searchClass&gt;"
  	//                   "&lt;upnp:searchClass includeDerived=\"1\"&gt;object.item.videoItem&lt;/upnp:searchClass");

    ret = strcatf(str,"&lt;upnp:writeStatus&gt;WRITABLE&lt;/upnp:writeStatus&gt;");
    ret = strcatf(str,"&lt;upnp:recordable&gt;0&lt;/upnp:recordable&gt;");
  }
  
  ret = strcatf(str, "&lt;dc:title&gt;%s&lt;/dc:title&gt;"
                     "&lt;upnp:writeStatus&gt;WRITABLE&lt;/upnp:writeStatus&gt;"
                     "&lt;upnp:class&gt;object.container&lt;/upnp:class&gt;",
                     psInputInfo->pbTitle);
                     //psInputInfo->pbTitle,psInputInfo->pbClass, psInputInfo->pbClass);
  
  
  //4. Minidlna Items
  //ret = strcatf(pbOutputStr, "&lt;upnp:storageUsed&gt;%s&lt;/upnp:storageUsed&gt;", (size ? size : "-1"));
  //ret = strcatf(pbOutputStr, "&lt;dc:creator&gt;%s&lt;/dc:creator&gt;", creator);
  //ret = strcatf(pbOutputStr, "&lt;upnp:genre&gt;%s&lt;/upnp:genre&gt;", genre);
  //ret = strcatf(pbOutputStr, "&lt;upnp:artist&gt;%s&lt;/upnp:artist&gt;", artist);
  
  //4. TZ85 items
  //<upnp:writeStatus>WRITABLE>/upnp:writeStatus>
  //<upnp:recordable>0>/upnp:recordable>
  if(psInputInfo->eContainerType != DLNA_WIFI_BROWSE_CONTAINER_TYPE_METADATA)  {
        #if 0
        if( passed_args->filter & FILTER_UPNP_ALBUMARTURI_DLNA_PROFILEID ) {
            ret = strcatf(str, "dlna:profileID=\"JPEG_TN\" xmlns:dlna=\"urn:schemas-dlna-org:metadata-1-0/\"");
        }
        #endif
        if(strcmp(psInputInfo->pbID, "0") != 0)  {
        	/*ret = strcatf(str, "&gt;http://%s:%d/DT%s&lt;/upnp:albumArtURI&gt;",
        	                   lan_addr[passed_args->iface].str, runtime_vars.port,psInputInfo->pbFilePath);*/
        
            
            EnumCustomizeTableIndex eSelectFileType = PANA_TABLE_INDEX_BROWSE_GROUPTYPE_STOPMOTION;
            switch(psInputInfo->eFileType)
            {
                case HCE_WIFI_DATABASE_FILE_TYPE_STOP_MOTION:
                    eSelectFileType = PANA_TABLE_INDEX_BROWSE_GROUPTYPE_STOPMOTION;
                    break;
                    
                case HCE_WIFI_DATABASE_FILE_TYPE_TIMELAPSE:
                    eSelectFileType = PANA_TABLE_INDEX_BROWSE_GROUPTYPE_TIMELAPSE;
                    break;
                case HCE_WIFI_DATABASE_FILE_TYPE_BURST:
                    eSelectFileType = PANA_TABLE_INDEX_BROWSE_GROUPTYPE_BURST;
                    break;
                default:
                    DLNA_Printf(E_DEBUG, L_GENERAL, "[PPD]Warning the type(%d) is wrong\n",psInputInfo->eFileType);
                    break;
            }    
            //Pana contain thumbnail URL
            ret = strcatf(str, "&lt;%s&gt;%s&lt;/%s&gt;",
                  GetDLNACustomizedTableString(PANA_TABLE_INDEX_BROWSE_TAG_GROUPTYPE),
                  GetDLNACustomizedTableString(eSelectFileType),
                  GetDLNACustomizedTableString(PANA_TABLE_INDEX_BROWSE_TAG_GROUPTYPE));
            ret = strcatf(str, "&lt;%s&gt;http://%s:%d/DT%s&lt;/%s&gt;",
                  GetDLNACustomizedTableString(PANA_TABLE_INDEX_BROWSE_TAG_GROUP_TN_URL),
                  lan_addr[passed_args->iface].str, runtime_vars.port, psInputInfo->pbFilePath,
                  GetDLNACustomizedTableString(PANA_TABLE_INDEX_BROWSE_TAG_GROUP_TN_URL));
            /*ret = strcatf(str, "&gt;http://%s:%d/%s.jpg&lt;/upnp:albumArtURI&gt;",
        	                   lan_addr[passed_args->iface].str, runtime_vars.port,psInputInfo->pbFilePath);*/
            
        }    
  }
  ret = strcatf(str, "&lt;/container&gt;");
  DLNA_Printf(E_DEBUG, L_GENERAL,"[PPD] Retrive contain done\n");
   
}    
typedef struct _UpnpBrowseResponseInputStruct_
{
    char pbObjectID[16];

}UpnpBrowseResponseInputStruct;
typedef struct _UpnpBrowseResponseOutputStruct_
{
    unsigned int ulTotalMatchNum;
    unsigned int ulCurrentReturnNum;
}UpnpBrowseResponseOutputStruct;

static void GetFileInformation(HceWiFiDatabaseFileType eFileType, Hce_WiFi_Command_State_MediaInfo_ResParaStruct sFileInfoRes, UpnpSoapBrowseItemStructure sItemInfo, struct Response *args)
{
    DPRINTF(E_DEBUG, L_GENERAL, "eFileType = 0x%02x\n", eFileType);
    if(eFileType != HCE_WIFI_DATABASE_FILE_TYPE_VIDEO) {
          DPRINTF(E_DEBUG, L_GENERAL, "sFileInfoRes.ResWidth = %lu, sFileInfoRes.ResHeight = %lu\n", sFileInfoRes.ResWidth, sFileInfoRes.ResHeight);
          if(sFileInfoRes.ResWidth > 4096 || sFileInfoRes.ResHeight > 4096)  {
              //strcpy(sItemInfo.pbFileType,"JPEG_LRG");
              DPRINTF(E_DEBUG, L_GENERAL, "Verify\n");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_ORG;
              AbilityDLNAAddResizedRes(&sItemInfo,args);

              strcpy(sItemInfo.pbFileType,"JPEG_TN");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
              sItemInfo.bIsResized      = 1;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
  
              memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
              strcpy(sItemInfo.pbFileType,"JPEG_LRG");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
              sItemInfo.bIsResized      = 0;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
          }
          else if(sFileInfoRes.ResWidth >= 1920 || sFileInfoRes.ResHeight >= 1080)  {
              strcpy(sItemInfo.pbFileType,"JPEG_LRG");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_ORG;
              sItemInfo.bIsResized      = 0;
              AbilityDLNAAddResizedRes(&sItemInfo,args);

              memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
              strcpy(sItemInfo.pbFileType,"JPEG_TN");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
              sItemInfo.bIsResized      = 1;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
  
              memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
              strcpy(sItemInfo.pbFileType,"JPEG_LRG");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG;
              sItemInfo.bIsResized      = 1;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
          }
          else if(sFileInfoRes.ResWidth >= 1024 || sFileInfoRes.ResHeight >= 768)  {
              strcpy(sItemInfo.pbFileType,"JPEG_LRG");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_ORG;
              sItemInfo.bIsResized      = 0;
              AbilityDLNAAddResizedRes(&sItemInfo,args);

              memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
              strcpy(sItemInfo.pbFileType,"JPEG_TN");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
              sItemInfo.bIsResized      = 1;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
  
              memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
              strcpy(sItemInfo.pbFileType,"JPEG_LRG");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG;
              sItemInfo.bIsResized      = 1;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
          }
          else if(sFileInfoRes.ResWidth >= 640 || sFileInfoRes.ResHeight >= 480)  {
              strcpy(sItemInfo.pbFileType,"JPEG_MED");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_ORG;
              sItemInfo.bIsResized      = 0;
              AbilityDLNAAddResizedRes(&sItemInfo,args);

              memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
              strcpy(sItemInfo.pbFileType,"JPEG_TN");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
              sItemInfo.bIsResized      = 1;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
  
              memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
              strcpy(sItemInfo.pbFileType,"JPEG_MED");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG;
              sItemInfo.bIsResized      = 1;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
          }
          else {
              strcpy(sItemInfo.pbFileType,"JPEG_SM");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_ORG;
              sItemInfo.bIsResized      = 0;
              AbilityDLNAAddResizedRes(&sItemInfo,args);

              memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
              strcpy(sItemInfo.pbFileType,"JPEG_TN");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
              sItemInfo.bIsResized      = 1;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
  
              memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
              strcpy(sItemInfo.pbFileType,"JPEG_SM");
              sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG;
              sItemInfo.bIsResized      = 1;
              AbilityDLNAAddResizedRes(&sItemInfo,args);
          }
       }
       //Viedo Process
       else  {
           if(sFileInfoRes.format == 0)  {
               sItemInfo.eVideoFormat = DLNA_WIFI_VIDEO_TYPE_MP4;
               if(sFileInfoRes.ResHeight == 1080)  {
                   if(sFileInfoRes.FrameRate == 60)  {
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_1080P_60;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 50)  {
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_1080P_50;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 30)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_HD_1080i_AAC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_1080P_30;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }

                   else if(sFileInfoRes.FrameRate == 25)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_HD_1080i_AAC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_1080P_25;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
               }
               else if(sFileInfoRes.ResHeight == 720)  {
                   if(sFileInfoRes.FrameRate == 60)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_HD_1080i_AAC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_720P_60;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 50)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_HD_1080i_AAC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_720P_50;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 30)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_HD_720p_AAC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_720P_30;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }

                   else if(sFileInfoRes.FrameRate == 25)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_HD_720p_AAC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_720P_25;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
               }
               else if(sFileInfoRes.ResHeight == 480)  {
                   if(sFileInfoRes.FrameRate == 30)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_HD_720p_AAC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_480P_30;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 25)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_HD_720p_AAC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_480P_25;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
               }
               else if(sFileInfoRes.ResHeight == 360)  {
                   if(sFileInfoRes.FrameRate == 30)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_SD_AAC_LC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_360P_30;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 25)  {
                       strcpy(sItemInfo.pbFileType,"AVC_MP4_MP_SD_AAC_LC");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_360P_25;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
               }
           }
           else if(sFileInfoRes.format == 1)  {
               sItemInfo.eVideoFormat = DLNA_WIFI_VIDEO_TYPE_TS;
               if(sFileInfoRes.ResHeight == 1080)  {
                   if(sFileInfoRes.FrameRate == 60)  {
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_1080P_60;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 50)  {
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_1080P_50;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 30)  {
                       strcpy(sItemInfo.pbFileType,"AVC_TS_HD_60_AC3_T");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_1080P_30;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }

                   else if(sFileInfoRes.FrameRate == 25)  {
                       strcpy(sItemInfo.pbFileType,"AVC_TS_HD_60_AC3_T");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_1080P_25;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
               }
               else if(sFileInfoRes.ResHeight == 720)  {
                   if(sFileInfoRes.FrameRate == 60)  {
                       strcpy(sItemInfo.pbFileType,"AVC_TS_HD_60_AC3_T");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_720P_60;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 50)  {
                       strcpy(sItemInfo.pbFileType,"AVC_TS_HD_60_AC3_T");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_720P_50;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
                   else if(sFileInfoRes.FrameRate == 30)  {
                       strcpy(sItemInfo.pbFileType,"AVC_TS_HD_60_AC3_T");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_720P_30;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }

                   else if(sFileInfoRes.FrameRate == 25)  {
                       strcpy(sItemInfo.pbFileType,"AVC_TS_HD_60_AC3_T");
                       sItemInfo.eVideoType       = DLNA_WIFI_VIDEO_TYPE_720P_25;
                       AbilityDLNAAddVideoRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_TN");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_TN;
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
           
                       memset(sItemInfo.pbFileType, 0, sizeof(sItemInfo.pbFileType));
                       strcpy(sItemInfo.pbFileType,"JPEG_SM");
                       sItemInfo.eFileSize       = DLNA_WIFI_FILE_SIZE_TYPE_LRG; 
                       AbilityDLNAAddResizedRes(&sItemInfo,args);
                   }
               }
           }
       }
}

static int UpnpSoapBrowseResponse(void *args,UpnpBrowseResponseInputStruct* psInputInfo,UpnpBrowseResponseOutputStruct *psOutputInfo )
{

  struct Response *passed_args = (struct Response *)args;
  struct string_s *str         = passed_args->str;
  int ret = 0;
  DLNA_Printf(E_DEBUG, L_GENERAL,"[PPD] DBG\n");
   
  int lCountI       = 0;
  int lHceResNumber = 0;
  Hce_WiFi_MsgID_Get_Group_Info_ReqParaStruct   sGroupInfoReq      = {{0}};
  Hce_WiFi_MsgID_Get_Group_Info_ResParaStruct   sGroupInfoResponse = {{0}};
  
  //----------------------------------------------------------------------------------------------------------------------
  //---------------------------------------Retrieve data from HCE---------------------------------------------------------
  //----------------------------------------------------------------------------------------------------------------------
  
  //4.!!@@## need to modify the input value
  sGroupInfoReq.eGroupType    = HCE_WIFI_GROUP_FINFO_FILE_TYPE_ALL;
  sGroupInfoReq.ulTotalNumber = 15;
  sGroupInfoReq.ulStartIndex  = 1;
  
  //3.!!@@## check parent ID and get corresponding type  
  HceDLNA_GeneralMessageToWiFi(HCE_WIFI_MESID_CGI_PLAYCMD_GET_GROUP_INFROMATION,
                           &sGroupInfoReq,
                           sizeof(Hce_WiFi_MsgID_Get_Group_Info_ReqParaStruct),
                           &sGroupInfoResponse,
                           sizeof(Hce_WiFi_MsgID_Get_Group_Info_ResParaStruct));
  lHceResNumber = sGroupInfoResponse.ulCurrentResponseNumber;
  
  //----------------------------------------------------------------------------------------------------------------------

    
  //------------------------------------------Construct response------------------------------------------------------
  for(lCountI = 0;lCountI < lHceResNumber; lCountI++)
  {
       //construct response data
     char*        pbClass        = "imageItem";
     char*        pbWriteStatus  = "WRITABLE";
     char         pbDetailID[64] = {0};
     char         tmpFolderIdx[4] = {0};
     char         tmpFileIdx[5] = {0};
     unsigned int ulTitleGroupIndex = sGroupInfoResponse.psGroupFileInfo[lCountI].ulGroupIndex;
     unsigned int ulTitleEntryIndex = sGroupInfoResponse.psGroupFileInfo[lCountI].ulGroupTotalEntry;
     memcpy(tmpFolderIdx,
        sGroupInfoResponse.psGroupFileInfo[lCountI].pubFileFullPath+strlen(sGroupInfoResponse.psGroupFileInfo[lCountI].pubFileFullPath)-11,
        3);
     memcpy(tmpFileIdx,
        sGroupInfoResponse.psGroupFileInfo[lCountI].pubFileFullPath+strlen(sGroupInfoResponse.psGroupFileInfo[lCountI].pubFileFullPath)-8,
        4);
     //4 !!@@## according to file type to set pbClass
     if(sGroupInfoResponse.psGroupFileInfo[lCountI].eFileType == HCE_WIFI_DATABASE_FILE_TYPE_TIMELAPSE  ||
        sGroupInfoResponse.psGroupFileInfo[lCountI].eFileType == HCE_WIFI_DATABASE_FILE_TYPE_BURST      ||
        sGroupInfoResponse.psGroupFileInfo[lCountI].eFileType == HCE_WIFI_DATABASE_FILE_TYPE_STOP_MOTION )  {
        //Container 
        DLNA_Printf(E_DEBUG, L_GENERAL,"[PPD] Container :%d\n",sGroupInfoResponse.psGroupFileInfo[lCountI].eFileType);
        UpnpSoapBrowseContainerStructure sInputInfo = {{0}};
        char tmpFileName[12] = {0};
        sInputInfo.eFileType = sGroupInfoResponse.psGroupFileInfo[lCountI].eFileType;
        sInputInfo.eContainerType = DLNA_WIFI_BROWSE_CONTAINER_TYPE_GROUP;
        snprintf(sInputInfo.pbClass,sizeof(sInputInfo.pbClass),"%s","imageItem");
        snprintf(sInputInfo.pbID,sizeof(sInputInfo.pbID),"%s%04dDIR",tmpFolderIdx,sGroupInfoResponse.psGroupFileInfo[lCountI].ulGroupTotalEntry);
        snprintf(sInputInfo.pbParent,sizeof(sInputInfo.pbParent),"%s","0");
        snprintf(sInputInfo.pbTitle,sizeof(sInputInfo.pbTitle),"%s-%s",tmpFolderIdx,tmpFileIdx);
        memcpy(sInputInfo.pbFilePath, 
            sGroupInfoResponse.psGroupFileInfo[lCountI].pubFileFullPath+strlen(sGroupInfoResponse.psGroupFileInfo[lCountI].pubFileFullPath)-11,
            11);
        sInputInfo.lChildCount = sGroupInfoResponse.psGroupFileInfo[lCountI].ulGroupTotalEntry;
        UpnpSoapConstructContainerData(args, &sInputInfo);
     }  
     else  {
       //Item
       // 1.Item id 2.ParentID 3.Retricted
       ret = strcatf(str, "&lt;item id=\"%s%s\" parentID=\"%s\" restricted=\"0\"&gt;",tmpFolderIdx,tmpFileIdx, psInputInfo->pbObjectID);
       // 4.dc:title 
       ret = strcatf(str, "&lt;dc:title&gt;%s-%s&lt;/dc:title&gt;",tmpFolderIdx,tmpFileIdx);
       #if 1
       //Panasonic
       //5.upnp:writeStatus
       ret = strcatf(str, "&lt;upnp:writeStatus&gt;%s&lt;/upnp:writeStatus&gt;",pbWriteStatus);
       //6.upnp:class name
       ret = strcatf(str, "&lt;upnp:class name=\"%s\"&gt;object.item.%s&lt;/upnp:class&gt;",pbClass, pbClass);
       #else
       //miniDLNA sample
       ret = strcatf(str, "&lt;upnp:class&gt;object.item.%s.photo&lt;/upnp:class&gt;", pbClass);
       ret = strcatf(str, "&lt;dc:creator&gt;Panasonic DMC-TZ85&lt;/dc:creator&gt;", pbClass);
       ret = strcatf(str, "&lt;dc:date&gt;2017-05-27T11:21:%d&lt;/dc:date&gt;",random()%60);
       #endif
       snprintf(pbDetailID,sizeof(pbDetailID),"%s%s",tmpFolderIdx,tmpFileIdx);
      
       //7.protocolInfo LRG
       //8.protocolInfo TN
       //9.protocolInfo MED
       //----------------Get File Attribute----------------//
       Hce_WiFi_Command_State_MediaInfo_ReqParaStruct sFileInfoReq = {0};       
       Hce_WiFi_Command_State_MediaInfo_ResParaStruct sFileInfoRes = {0};

       sFileInfoReq.RetrieveMethod = CAMERA_DATABASE_RETRIEVE_BY_NAME;
       memcpy(sFileInfoReq.FilePathString, sGroupInfoResponse.psGroupFileInfo[lCountI].pubFileFullPath+3, strlen(sGroupInfoResponse.psGroupFileInfo[lCountI].pubFileFullPath)-3);
       DPRINTF(E_DEBUG, L_GENERAL, "sFileInfoReq.FilePathString: %s\n", sFileInfoReq.FilePathString);
       
       HceDLNA_GeneralMessageToWiFi(HCE_WIFI_MSGID_CGI_PLAYCMD_GET_FILE_INFORMATION,
                           &sFileInfoReq,
                           sizeof(Hce_WiFi_Command_State_MediaInfo_ReqParaStruct),
                           &sFileInfoRes,
                           sizeof(Hce_WiFi_Command_State_MediaInfo_ResParaStruct));

       UpnpSoapBrowseItemStructure sItemInfo = {{0}};
       sItemInfo.pbFileFullPath = sFileInfoReq.FilePathString;
       snprintf(sItemInfo.pbResolution, sizeof(sItemInfo.pbResolution), "%lux%lu", sFileInfoRes.ResWidth, sFileInfoRes.ResHeight);
       sItemInfo.eVideoFormat = sFileInfoRes.format;

       DPRINTF(E_DEBUG, L_GENERAL, "File Type: 0x%02x\n", sGroupInfoResponse.psGroupFileInfo[lCountI].eFileType);
       GetFileInformation(sGroupInfoResponse.psGroupFileInfo[lCountI].eFileType, sFileInfoRes, sItemInfo, args);

       ret = strcatf(str, "&lt;/item&gt;");
     }  
     passed_args->returned ++ ;
  }  
    
  psOutputInfo->ulTotalMatchNum    = sGroupInfoResponse.ulTotalMatchedNumber;
  psOutputInfo->ulCurrentReturnNum = sGroupInfoResponse.ulCurrentResponseNumber;

  return 0;
}

static int UpnpSoapContainerBrowseResponse(void *args,UpnpBrowseResponseInputStruct* psInputInfo,UpnpBrowseResponseOutputStruct *psOutputInfo, int StartingIdx )
{

  struct Response *passed_args = (struct Response *)args;
  struct string_s *str         = passed_args->str;
  int ret = 0;
  DLNA_Printf(E_DEBUG, L_GENERAL,"[PPD] DBG\n");
   
  int lCountI       = 0;
  int lHceResNumber = 0;
  char tmpGroupIdx[4] = {0};
  Hce_WiFi_MsgID_Get_Group_Entry_Info_ReqParaStruct   sGroupEntryInfoReq      = {{0}};
  Hce_WiFi_MsgID_Get_Group_Entry_Info_ResParaStruct   sGroupEntryInfoResponse = {{0}};
  //-----Parse Group Index and Total Number
  memcpy(tmpGroupIdx, psInputInfo->pbObjectID, 3);
  //----------------------------------------------------------------------------------------------------------------------
  //---------------------------------------Retrieve data from HCE---------------------------------------------------------
  //----------------------------------------------------------------------------------------------------------------------
  sGroupEntryInfoReq.ulGroupIndex = atoi(tmpGroupIdx);
  sGroupEntryInfoReq.ulTotalNumber = passed_args->requested;
  sGroupEntryInfoReq.ulStartingIndex = StartingIdx;
  DLNA_Printf(E_DEBUG, L_GENERAL,"GroupIndex = %lu\n", sGroupEntryInfoReq.ulGroupIndex);
  DLNA_Printf(E_DEBUG, L_GENERAL,"RequestNum = %lu\n", sGroupEntryInfoReq.ulTotalNumber);
  DLNA_Printf(E_DEBUG, L_GENERAL,"StartingIndex = %lu\n", sGroupEntryInfoReq.ulStartingIndex);
  
  //3.!!@@## check parent ID and get corresponding type  
  HceDLNA_GeneralMessageToWiFi(HCE_WIFI_MESID_CGI_PLAYCMD_GET_GROUP_ENTRY_INFROMATION,
                           &sGroupEntryInfoReq,
                           sizeof(Hce_WiFi_MsgID_Get_Group_Entry_Info_ReqParaStruct),
                           &sGroupEntryInfoResponse,
                           sizeof(Hce_WiFi_MsgID_Get_Group_Entry_Info_ResParaStruct));
  DLNA_Printf(E_DEBUG, L_GENERAL,"MSG come back\n");
  lHceResNumber = sGroupEntryInfoResponse.ulCurrentResponseNumber;
  
  //----------------------------------------------------------------------------------------------------------------------

    
  //------------------------------------------Construct response------------------------------------------------------
  for(lCountI = 0;lCountI < lHceResNumber; lCountI++)
  {
       //construct response data
     char*        pbClass        = "imageItem";
     char*        pbWriteStatus  = "WRITABLE";
     char         pbDetailID[64] = {0};
     char         tmpFolderIdx[4] = {0};
     char         tmpFileIdx[5] = {0};
     unsigned int ulTitleGroupIndex = 100;
     unsigned int ulTitleEntryIndex = sGroupEntryInfoResponse.psEntryinfo[lCountI].ulFileIndex;
     
     //4 !!@@## according to file type to set pbClass
     
     //Item
     // 1.Item id 2.ParentID 3.Retricted
     //ret = strcatf(str, "&lt;item id=\"%03d%04d\" parentID=\"%s\" restricted=\"0\"&gt;",ulTitleGroupIndex,ulTitleEntryIndex, psInputInfo->pbObjectID);
     ret = strcatf(str, "&lt;item id=\"1000001\" parentID=\"%s\" restricted=\"0\"&gt;",psInputInfo->pbObjectID);
     // 4.dc:title 
     //ret = strcatf(str, "&lt;dc:title&gt;%03d-%04d&lt;/dc:title&gt;",ulTitleGroupIndex,ulTitleEntryIndex);
     ret = strcatf(str, "&lt;dc:title&gt;100-0001&lt;/dc:title&gt;");
     #if 1
     //Panasonic
     //5.upnp:writeStatus
     ret = strcatf(str, "&lt;upnp:writeStatus&gt;%s&lt;/upnp:writeStatus&gt;",pbWriteStatus);
     //6.upnp:class name
     ret = strcatf(str, "&lt;upnp:class name=\"%s\"&gt;object.item.%s&lt;/upnp:class&gt;",pbClass, pbClass);
     #else
     //miniDLNA sample
     ret = strcatf(str, "&lt;upnp:class&gt;object.item.%s.photo&lt;/upnp:class&gt;", pbClass);
     ret = strcatf(str, "&lt;dc:creator&gt;Panasonic DMC-TZ85&lt;/dc:creator&gt;", pbClass);
     ret = strcatf(str, "&lt;dc:date&gt;2017-05-27T11:21:%d&lt;/dc:date&gt;",random()%60);
     #endif
     snprintf(pbDetailID,sizeof(pbDetailID),"%03d%03d",ulTitleGroupIndex,ulTitleEntryIndex);
    
     //7.protocolInfo LRG
     //8.protocolInfo TN
     //9.protocolInfo MED
     //----------------Get File Attribute----------------//
     Hce_WiFi_Command_State_MediaInfo_ReqParaStruct sFileInfoReq = {0};       
     Hce_WiFi_Command_State_MediaInfo_ResParaStruct sFileInfoRes = {0};

     sFileInfoReq.RetrieveMethod = CAMERA_DATABASE_RETRIEVE_BY_NAME;
     memcpy(sFileInfoReq.FilePathString, sGroupEntryInfoResponse.psEntryinfo[lCountI].pubFileFullPath, HCE_WIFI_CAMERA_DATAPATH_STRLEN);
     DPRINTF(E_DEBUG, L_GENERAL, "sFileInfoReq.FilePathString: %s\n", sFileInfoReq.FilePathString);
     
     HceDLNA_GeneralMessageToWiFi(HCE_WIFI_MSGID_CGI_PLAYCMD_GET_FILE_INFORMATION,
                         &sFileInfoReq,
                         sizeof(Hce_WiFi_Command_State_MediaInfo_ReqParaStruct),
                         &sFileInfoRes,
                         sizeof(Hce_WiFi_Command_State_MediaInfo_ResParaStruct));

     UpnpSoapBrowseItemStructure sItemInfo = {{0}};
     sItemInfo.pbFileFullPath = sFileInfoReq.FilePathString;
     snprintf(sItemInfo.pbResolution, sizeof(sItemInfo.pbResolution), "%lux%lu", sFileInfoRes.ResWidth, sFileInfoRes.ResHeight);
     sItemInfo.eVideoFormat = sFileInfoRes.format;

     GetFileInformation(HCE_WIFI_DATABASE_FILE_TYPE_SINGLE_PHOTO, sFileInfoRes, sItemInfo, args);     

       ret = strcatf(str, "&lt;/item&gt;");
       
     passed_args->returned ++ ;
  }  
    
  psOutputInfo->ulTotalMatchNum    = sGroupEntryInfoReq.ulTotalNumber;
  psOutputInfo->ulCurrentReturnNum = sGroupEntryInfoReq.ulTotalNumber;

  return 0;
}

#endif
static int
UpnpSoapCallback(void *args, int argc, char **argv, char **azColName)
{
	struct Response *passed_args = (struct Response *)args;
	char *id = argv[0], *parent = argv[1], *refID = argv[2], *detailID = argv[3], *class = argv[4], *size = argv[5], *title = argv[6],
	     *duration = argv[7], *bitrate = argv[8], *sampleFrequency = argv[9], *artist = argv[10], *album = argv[11],
	     *genre = argv[12], *comment = argv[13], *nrAudioChannels = argv[14], *track = argv[15], *date = argv[16], *resolution = argv[17],
	     *tn = argv[18], *creator = argv[19], *dlna_pn = argv[20], *mime = argv[21], *album_art = argv[22], *rotate = argv[23];
	char dlna_buf[128];
	const char *ext;
	struct string_s *str = passed_args->str;
	int ret = 0;

	/* Make sure we have at least 8KB left of allocated memory to finish the response. */
	if( str->off > (str->size - 8192) )
	{
#if MAX_RESPONSE_SIZE > 0
		if( (str->size+DEFAULT_RESP_SIZE) <= MAX_RESPONSE_SIZE )
		{
#endif
			str->data = realloc(str->data, (str->size+DEFAULT_RESP_SIZE));
			if( str->data )
			{
				str->size += DEFAULT_RESP_SIZE;
				DPRINTF(E_DEBUG, L_HTTP, "UPnP SOAP response enlarged to %lu. [%d results so far]\n",
					(unsigned long)str->size, passed_args->returned);
			}
			else
			{
				DPRINTF(E_ERROR, L_HTTP, "UPnP SOAP response was too big, and realloc failed!\n");
				return -1;
			}
#if MAX_RESPONSE_SIZE > 0
		}
		else
		{
			DPRINTF(E_ERROR, L_HTTP, "UPnP SOAP response cut short, to not exceed the max response size [%lld]!\n", (long long int)MAX_RESPONSE_SIZE);
			return -1;
		}
#endif
	}

//-----------------------------------------------------------------------------------------    
	passed_args->returned++;
	passed_args->flags &= ~RESPONSE_FLAGS;

	if( strncmp(class, "item", 4) == 0 )
	{
		uint32_t dlna_flags = DLNA_FLAG_DLNA_V1_5|DLNA_FLAG_HTTP_STALLING|DLNA_FLAG_TM_B;
		char *alt_title = NULL;
		/* We may need special handling for certain MIME types */
 //--------------------------------------------------------------------------------------------------------------------------------
 //----------------------------------------Response with video type----------------------------------------------------------------
 //--------------------------------------------------------------------------------------------------------------------------------       
		if( *mime == 'v' )
		{
			dlna_flags |= DLNA_FLAG_TM_S;
			if( passed_args->flags & FLAG_MIME_AVI_DIVX )
			{
				if( strcmp(mime, "video/x-msvideo") == 0 )
				{
					if( creator )
						strcpy(mime+6, "divx");
					else
						strcpy(mime+6, "avi");
				}
			}
			else if( passed_args->flags & FLAG_MIME_AVI_AVI )
			{
				if( strcmp(mime, "video/x-msvideo") == 0 )
				{
					strcpy(mime+6, "avi");
				}
			}
			else if( passed_args->client == EFreeBox && dlna_pn )
			{
				if( strncmp(dlna_pn, "AVC_TS", 6) == 0 ||
				    strncmp(dlna_pn, "MPEG_TS", 7) == 0 )
				{
					strcpy(mime+6, "mp2t");
				}
			}
			if( !(passed_args->flags & FLAG_DLNA) )
			{
				if( strcmp(mime+6, "vnd.dlna.mpeg-tts") == 0 )
				{
					strcpy(mime+6, "mpeg");
				}
			}
			if( (passed_args->flags & FLAG_CAPTION_RES) ||
			    (passed_args->filter & (FILTER_SEC_CAPTION_INFO_EX|FILTER_PV_SUBTITLE)) )
			{
				if( sql_get_int_field(db, "SELECT ID from CAPTIONS where ID = '%s'", detailID) > 0 )
					passed_args->flags |= FLAG_HAS_CAPTIONS;
			}
			/* From what I read, Samsung TV's expect a [wrong] MIME type of x-mkv. */
			if( passed_args->flags & FLAG_SAMSUNG )
			{
				if( strcmp(mime+6, "x-matroska") == 0 )
				{
					strcpy(mime+8, "mkv");
				}
			}
			/* LG hack: subtitles won't get used unless dc:title contains a dot. */
			else if( passed_args->client == ELGDevice && (passed_args->flags & FLAG_HAS_CAPTIONS) )
			{
				ret = asprintf(&alt_title, "%s.", title);
				if( ret > 0 )
					title = alt_title;
				else
					alt_title = NULL;
			}
			/* Asus OPlay reboots with titles longer than 23 characters with some file types. */
			else if( passed_args->client == EAsusOPlay && (passed_args->flags & FLAG_HAS_CAPTIONS) )
			{
				if( strlen(title) > 23 )
					title[23] = '\0';
			}
		}

 //--------------------------------------------------------------------------------------------------------------------------------
 //----------------------------------------Response with Audio type----------------------------------------------------------------
 //--------------------------------------------------------------------------------------------------------------------------------         
		else if( *mime == 'a' )
		{
			dlna_flags |= DLNA_FLAG_TM_S;
			if( strcmp(mime+6, "x-flac") == 0 )
			{
				if( passed_args->flags & FLAG_MIME_FLAC_FLAC )
				{
					strcpy(mime+6, "flac");
				}
			}
			else if( strcmp(mime+6, "x-wav") == 0 )
			{
				if( passed_args->flags & FLAG_MIME_WAV_WAV )
				{
					strcpy(mime+6, "wav");
				}
			}
		}

 //--------------------------------------------------------------------------------------------------------------------------------
 //----------------------------------------Response else----------------------------------------------------------------
 //--------------------------------------------------------------------------------------------------------------------------------         
		else
			dlna_flags |= DLNA_FLAG_TM_I;

 //--------------------------------------------------------------------------------------------------------------------------------   
 //--------------------------------------------------------------------------------------------------------------------------------   
 //--------------------------------------------------------------------------------------------------------------------------------   
		if( dlna_pn )
			snprintf(dlna_buf, sizeof(dlna_buf), "DLNA.ORG_PN=%s;"
			                                     "DLNA.ORG_OP=01;"
			                                     "DLNA.ORG_CI=0;"
			                                     "DLNA.ORG_FLAGS=%08X%024X",
			                                     dlna_pn, dlna_flags, 0);
		else if( passed_args->flags & FLAG_DLNA )
			snprintf(dlna_buf, sizeof(dlna_buf), "DLNA.ORG_OP=01;"
			                                     "DLNA.ORG_CI=0;"
			                                     "DLNA.ORG_FLAGS=%08X%024X",
			                                     dlna_flags, 0);
		else
			strcpy(dlna_buf, "*");

		ret = strcatf(str, "&lt;item id=\"%s\" parentID=\"%s\" restricted=\"1\"", id, parent);
		if( refID && (passed_args->filter & FILTER_REFID) ) {
			ret = strcatf(str, " refID=\"%s\"", refID);
		}
        
	    DPRINTF(E_DEBUG, L_GENERAL, "[PPD][%s %d] title :%s comment:%s\n",__FUNCTION__,__LINE__,title,comment);
		ret = strcatf(str, "&gt;"
		                   "&lt;dc:title&gt;%s&lt;/dc:title&gt;"
		                   "&lt;upnp:class&gt;object.%s&lt;/upnp:class&gt;",
		                   title, class);
		if( comment && (passed_args->filter & FILTER_DC_DESCRIPTION) ) {
	        DPRINTF(E_DEBUG, L_GENERAL, "[PPD][%s %d] comment:%s\n",__FUNCTION__,__LINE__,comment);
			ret = strcatf(str, "&lt;dc:description&gt;%.384s&lt;/dc:description&gt;", comment);
		}
		if( creator && (passed_args->filter & FILTER_DC_CREATOR) ) {
			ret = strcatf(str, "&lt;dc:creator&gt;%s&lt;/dc:creator&gt;", creator);
		}
		if( date && (passed_args->filter & FILTER_DC_DATE) ) {
			ret = strcatf(str, "&lt;dc:date&gt;%s&lt;/dc:date&gt;", date);
		}
		if( passed_args->filter & FILTER_SEC_DCM_INFO ) {
			/* Get bookmark */
			ret = strcatf(str, "&lt;sec:dcmInfo&gt;CREATIONDATE=0,FOLDER=%s,BM=%d&lt;/sec:dcmInfo&gt;",
			              title, sql_get_int_field(db, "SELECT SEC from BOOKMARKS where ID = '%s'", detailID));
		}
		if( artist ) {
			if( (*mime == 'v') && (passed_args->filter & FILTER_UPNP_ACTOR) ) {
				ret = strcatf(str, "&lt;upnp:actor&gt;%s&lt;/upnp:actor&gt;", artist);
			}
			if( passed_args->filter & FILTER_UPNP_ARTIST ) {
				ret = strcatf(str, "&lt;upnp:artist&gt;%s&lt;/upnp:artist&gt;", artist);
			}
		}
		if( album && (passed_args->filter & FILTER_UPNP_ALBUM) ) {
			ret = strcatf(str, "&lt;upnp:album&gt;%s&lt;/upnp:album&gt;", album);
		}
		if( genre && (passed_args->filter & FILTER_UPNP_GENRE) ) {
			ret = strcatf(str, "&lt;upnp:genre&gt;%s&lt;/upnp:genre&gt;", genre);
		}
//-----------------------------------------------------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------------------------------------------------        
		if( strncmp(id, MUSIC_PLIST_ID, strlen(MUSIC_PLIST_ID)) == 0 ) {
			track = strrchr(id, '$')+1;
		}
		if( NON_ZERO(track) && (passed_args->filter & FILTER_UPNP_ORIGINALTRACKNUMBER) ) {
			ret = strcatf(str, "&lt;upnp:originalTrackNumber&gt;%s&lt;/upnp:originalTrackNumber&gt;", track);
		}
		if( passed_args->filter & FILTER_RES ) {
			ext = mime_to_ext(mime);
			add_res(size, duration, bitrate, sampleFrequency, nrAudioChannels,
			        resolution, dlna_buf, mime, detailID, ext, passed_args);
			if( *mime == 'i' ) {
				int srcw, srch;
				if( resolution && (sscanf(resolution, "%6dx%6d", &srcw, &srch) == 2) )
				{
					if( srcw > 4096 || srch > 4096 )
						add_resized_res(srcw, srch, 4096, 4096, "JPEG_LRG", detailID, passed_args);
					if( srcw > 1024 || srch > 768 )
						add_resized_res(srcw, srch, 1024, 768, "JPEG_MED", detailID, passed_args);
					if( srcw > 640 || srch > 480 )
						add_resized_res(srcw, srch, 640, 480, "JPEG_SM", detailID, passed_args);
				}
				if( !(passed_args->flags & FLAG_RESIZE_THUMBS) && NON_ZERO(tn) && IS_ZERO(rotate) ) {
	                DPRINTF(E_DEBUG, L_GENERAL, "[PPD][%s %d] \n",__FUNCTION__,__LINE__);
					ret = strcatf(str, "&lt;res protocolInfo=\"http-get:*:%s:%s\"&gt;"
					                   "http://%s:%d/Thumbnails/%s.jpg"
					                   "&lt;/res&gt;",
					                   mime, "DLNA.ORG_PN=JPEG_TN;DLNA.ORG_CI=1", lan_addr[passed_args->iface].str,
					                   runtime_vars.port, detailID);
				}
				else
					add_resized_res(srcw, srch, 160, 160, "JPEG_TN", detailID, passed_args);
			}
			else if( *mime == 'v' ) {
				switch( passed_args->client ) {
				case EToshibaTV:
					if( dlna_pn &&
					    (strncmp(dlna_pn, "MPEG_TS_HD_NA", 13) == 0 ||
					     strncmp(dlna_pn, "MPEG_TS_SD_NA", 13) == 0 ||
					     strncmp(dlna_pn, "AVC_TS_MP_HD_AC3", 16) == 0 ||
					     strncmp(dlna_pn, "AVC_TS_HP_HD_AC3", 16) == 0))
					{
						sprintf(dlna_buf, "DLNA.ORG_PN=%s;DLNA.ORG_OP=01;DLNA.ORG_CI=1", "MPEG_PS_NTSC");
						add_res(size, duration, bitrate, sampleFrequency, nrAudioChannels,
						        resolution, dlna_buf, mime, detailID, ext, passed_args);
					}
					break;
				case ESonyBDP:
					if( dlna_pn &&
					    (strncmp(dlna_pn, "AVC_TS", 6) == 0 ||
					     strncmp(dlna_pn, "MPEG_TS", 7) == 0) )
					{
						if( strncmp(dlna_pn, "MPEG_TS_SD_NA", 13) != 0 )
						{
							sprintf(dlna_buf, "DLNA.ORG_PN=%s;DLNA.ORG_OP=01;DLNA.ORG_CI=1", "MPEG_TS_SD_NA");
							add_res(size, duration, bitrate, sampleFrequency, nrAudioChannels,
							        resolution, dlna_buf, mime, detailID, ext, passed_args);
						}
						if( strncmp(dlna_pn, "MPEG_TS_SD_EU", 13) != 0 )
						{
							sprintf(dlna_buf, "DLNA.ORG_PN=%s;DLNA.ORG_OP=01;DLNA.ORG_CI=1", "MPEG_TS_SD_EU");
							add_res(size, duration, bitrate, sampleFrequency, nrAudioChannels,
							        resolution, dlna_buf, mime, detailID, ext, passed_args);
						}
					}
					else if( (dlna_pn &&
					          (strncmp(dlna_pn, "AVC_MP4", 7) == 0 ||
					           strncmp(dlna_pn, "MPEG4_P2_MP4", 12) == 0)) ||
					         strcmp(mime+6, "x-matroska") == 0 ||
					         strcmp(mime+6, "x-msvideo") == 0 ||
					         strcmp(mime+6, "mpeg") == 0 )
					{
						strcpy(mime+6, "avi");
						if( !dlna_pn || strncmp(dlna_pn, "MPEG_PS_NTSC", 12) != 0 )
						{
							sprintf(dlna_buf, "DLNA.ORG_PN=%s;DLNA.ORG_OP=01;DLNA.ORG_CI=1", "MPEG_PS_NTSC");
							add_res(size, duration, bitrate, sampleFrequency, nrAudioChannels,
						        	resolution, dlna_buf, mime, detailID, ext, passed_args);
						}
						if( !dlna_pn || strncmp(dlna_pn, "MPEG_PS_PAL", 11) != 0 )
						{
							sprintf(dlna_buf, "DLNA.ORG_PN=%s;DLNA.ORG_OP=01;DLNA.ORG_CI=1", "MPEG_PS_PAL");
							add_res(size, duration, bitrate, sampleFrequency, nrAudioChannels,
						        	resolution, dlna_buf, mime, detailID, ext, passed_args);
						}
					}
					break;
				case ESonyBravia:
					/* BRAVIA KDL-##*X### series TVs do natively support AVC/AC3 in TS, but
					   require profile to be renamed (applies to _T and _ISO variants also) */
					if( dlna_pn &&
					    (strncmp(dlna_pn, "AVC_TS_MP_SD_AC3", 16) == 0 ||
					     strncmp(dlna_pn, "AVC_TS_MP_HD_AC3", 16) == 0 ||
					     strncmp(dlna_pn, "AVC_TS_HP_HD_AC3", 16) == 0))
					{
					        sprintf(dlna_buf, "DLNA.ORG_PN=AVC_TS_HD_50_AC3%s", dlna_pn + 16);
						add_res(size, duration, bitrate, sampleFrequency, nrAudioChannels,
						        resolution, dlna_buf, mime, detailID, ext, passed_args);
					}
					break;
				case ESamsungSeriesCDE:
				case ELGDevice:
				case ELGNetCastDevice:
				case EAsusOPlay:
				default:
					if( passed_args->flags & FLAG_HAS_CAPTIONS )
					{
						if( passed_args->flags & FLAG_CAPTION_RES )  {
	                        DPRINTF(E_DEBUG, L_GENERAL, "[PPD][%s %d] \n",__FUNCTION__,__LINE__);
							ret = strcatf(str, "&lt;res protocolInfo=\"http-get:*:text/srt:*\"&gt;"
									     "http://%s:%d/Captions/%s.srt"
									   "&lt;/res&gt;",
									   lan_addr[passed_args->iface].str, runtime_vars.port, detailID);


                        }    
						if( passed_args->filter & FILTER_SEC_CAPTION_INFO_EX )   {
							ret = strcatf(str, "&lt;sec:CaptionInfoEx sec:type=\"srt\"&gt;"
							                     "http://%s:%d/Captions/%s.srt"
							                   "&lt;/sec:CaptionInfoEx&gt;",
							                   lan_addr[passed_args->iface].str, runtime_vars.port, detailID);

                        }    
					}
					free(alt_title);
					break;
				}
			}
		}
		if( NON_ZERO(album_art) )
		{
			/* Video and audio album art is handled differently */
			if( *mime == 'v' && (passed_args->filter & FILTER_RES) && !(passed_args->flags & FLAG_MS_PFS) ) {
	            DPRINTF(E_DEBUG, L_GENERAL, "[PPD][%s %d] %s\n",__FUNCTION__,__LINE__,album_art);
				ret = strcatf(str, "&lt;res protocolInfo=\"http-get:*:image/jpeg:DLNA.ORG_PN=JPEG_TN\"&gt;"
				                   "http://%s:%d/AlbumArt/%s-%s.jpg"
				                   "&lt;/res&gt;",
				                   lan_addr[passed_args->iface].str, runtime_vars.port, album_art, detailID);
			} else if( passed_args->filter & FILTER_UPNP_ALBUMARTURI ) {
				ret = strcatf(str, "&lt;upnp:albumArtURI");
				if( passed_args->filter & FILTER_UPNP_ALBUMARTURI_DLNA_PROFILEID ) {
					ret = strcatf(str, " dlna:profileID=\"JPEG_TN\" xmlns:dlna=\"urn:schemas-dlna-org:metadata-1-0/\"");
				}
				ret = strcatf(str, "&gt;http://%s:%d/AlbumArt/%s-%s.jpg&lt;/upnp:albumArtURI&gt;",
				                   lan_addr[passed_args->iface].str, runtime_vars.port, album_art, detailID);
			}
		}
		if( (passed_args->flags & FLAG_MS_PFS) && *mime == 'i' ) {
			if( passed_args->client == EMediaRoom && !album )
				ret = strcatf(str, "&lt;upnp:album&gt;%s&lt;/upnp:album&gt;", "[No Keywords]");

			/* EVA2000 doesn't seem to handle embedded thumbnails */
			if( !(passed_args->flags & FLAG_RESIZE_THUMBS) && NON_ZERO(tn) && IS_ZERO(rotate) ) {
				ret = strcatf(str, "&lt;upnp:albumArtURI&gt;"
				                   "http://%s:%d/Thumbnails/%s.jpg"
				                   "&lt;/upnp:albumArtURI&gt;",
				                   lan_addr[passed_args->iface].str, runtime_vars.port, detailID);
			} else {
				ret = strcatf(str, "&lt;upnp:albumArtURI&gt;"
				                   "http://%s:%d/Resized/%s.jpg?width=160,height=160"
				                   "&lt;/upnp:albumArtURI&gt;",
				                   lan_addr[passed_args->iface].str, runtime_vars.port, detailID);
			}
		}
		ret = strcatf(str, "&lt;/item&gt;");
	}
	else if( strncmp(class, "container", 9) == 0 )
	{
	    DPRINTF(E_DEBUG, L_GENERAL, "[PPD][%s %d] id : %s  parent : %s\n",__FUNCTION__,__LINE__,id, parent);
		ret = strcatf(str, "&lt;container id=\"%s\" parentID=\"%s\" restricted=\"1\" ", id, parent);
		if( passed_args->filter & FILTER_SEARCHABLE ) {
			ret = strcatf(str, "searchable=\"%d\" ", check_magic_container(id, passed_args->flags) ? 0 : 1);
		}
		if( passed_args->filter & FILTER_CHILDCOUNT ) {
			ret = strcatf(str, "childCount=\"%d\"", get_child_count(id, check_magic_container(id, passed_args->flags)));
		}
		/* If the client calls for BrowseMetadata on root, we have to include our "upnp:searchClass"'s, unless they're filtered out */
		if( passed_args->requested == 1 && strcmp(id, "0") == 0 && (passed_args->filter & FILTER_UPNP_SEARCHCLASS) ) {
			ret = strcatf(str, "&gt;"
			                   "&lt;upnp:searchClass includeDerived=\"1\"&gt;object.item.audioItem&lt;/upnp:searchClass&gt;"
			                   "&lt;upnp:searchClass includeDerived=\"1\"&gt;object.item.imageItem&lt;/upnp:searchClass&gt;"
			                   "&lt;upnp:searchClass includeDerived=\"1\"&gt;object.item.videoItem&lt;/upnp:searchClass");
		}
		ret = strcatf(str, "&gt;"
		                   "&lt;dc:title&gt;%s&lt;/dc:title&gt;"
		                   "&lt;upnp:class&gt;object.%s&lt;/upnp:class&gt;",
		                   title, class);
		if( (passed_args->filter & FILTER_UPNP_STORAGEUSED) || strcmp(class+10, "storageFolder") == 0 ) {
			/* TODO: Implement real folder size tracking */
			ret = strcatf(str, "&lt;upnp:storageUsed&gt;%s&lt;/upnp:storageUsed&gt;", (size ? size : "-1"));
		}
		if( creator && (passed_args->filter & FILTER_DC_CREATOR) ) {
			ret = strcatf(str, "&lt;dc:creator&gt;%s&lt;/dc:creator&gt;", creator);
		}
		if( genre && (passed_args->filter & FILTER_UPNP_GENRE) ) {
			ret = strcatf(str, "&lt;upnp:genre&gt;%s&lt;/upnp:genre&gt;", genre);
		}
		if( artist && (passed_args->filter & FILTER_UPNP_ARTIST) ) {
			ret = strcatf(str, "&lt;upnp:artist&gt;%s&lt;/upnp:artist&gt;", artist);
		}
		if( NON_ZERO(album_art) && (passed_args->filter & FILTER_UPNP_ALBUMARTURI) ) {
			ret = strcatf(str, "&lt;upnp:albumArtURI ");
			if( passed_args->filter & FILTER_UPNP_ALBUMARTURI_DLNA_PROFILEID ) {
				ret = strcatf(str, "dlna:profileID=\"JPEG_TN\" xmlns:dlna=\"urn:schemas-dlna-org:metadata-1-0/\"");
			}
			ret = strcatf(str, "&gt;http://%s:%d/AlbumArt/%s-%s.jpg&lt;/upnp:albumArtURI&gt;",
			                   lan_addr[passed_args->iface].str, runtime_vars.port, album_art, detailID);
		}
		if( passed_args->filter & FILTER_AV_MEDIA_CLASS ) {
			char class;
			if( strncmp(id, MUSIC_ID, sizeof(MUSIC_ID)) == 0 )
				class = 'M';
			else if( strncmp(id, VIDEO_ID, sizeof(VIDEO_ID)) == 0 )
				class = 'V';
			else if( strncmp(id, IMAGE_ID, sizeof(IMAGE_ID)) == 0 )
				class = 'P';
			else
				class = 0;
			if( class )
				ret = strcatf(str, "&lt;av:mediaClass xmlns:av=\"urn:schemas-sony-com:av\"&gt;"
				                    "%c&lt;/av:mediaClass&gt;", class);
		}
		ret = strcatf(str, "&lt;/container&gt;");
	}

	return 0;
}

#ifdef ABILITY_MODIFICATION
static int ExamFileExist(char* FileID, char* FullPath)
{
    char TmpFullPath[64] = {0};
    struct stat fstat = {0};
    int ExamResult = 0;
    char* ToSysPath;
    strcat(TmpFullPath, "/media/ampsd/DCIM/");
    strncat(TmpFullPath, FileID, 3);
    strcat(TmpFullPath, "_PANA/P");
    strcat(TmpFullPath, FileID);
    strcat(TmpFullPath, ".JPG");
    DPRINTF(E_DEBUG, L_HTTP, "TmpFullPath: %s\n", TmpFullPath);
    if((stat(TmpFullPath, &fstat)) != -1)  {
        ToSysPath = strstr(TmpFullPath, "/DCIM");
        memcpy(FullPath, ToSysPath, strlen(ToSysPath));
        return 0;
    }
    memcpy(FullPath + (strlen(FullPath) - 3), "MP4", strlen("MP4"));
    if((stat(TmpFullPath, &fstat)) != -1)  {
        ToSysPath = strstr(TmpFullPath, "/DCIM");
        memcpy(FullPath, ToSysPath, strlen(ToSysPath));
        return 0;
    }
    memcpy(FullPath + (strlen(FullPath) - 3), "MOV", strlen("MOV"));
    if((stat(TmpFullPath, &fstat)) != -1)  {
        ToSysPath = strstr(TmpFullPath, "/DCIM");
        memcpy(FullPath, ToSysPath, strlen(ToSysPath));
        return 0;
    }
    return -1;
    
}
static void ExecDestroyObject(struct upnphttp * h, const char * action)
{
    
	int ret = 0;
    static const char resp0[] = 
          "<u:DestroyObjectResponse xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\">"
          "</u:DestroyObjectResponse>";

	struct string_s str;
    char FileFullPath[64] = {0};
    
	memset(&str, 0, sizeof(str));
	str.data = malloc(DLNA_SOAP_NORMAL_RESP_SIZE);
    memset(str.data,0,DLNA_SOAP_NORMAL_RESP_SIZE);
	str.size = DLNA_SOAP_NORMAL_RESP_SIZE;

    
    char *ObjectID = NULL;
	struct NameValueParserData data;
    Hce_WiFi_Command_Execute_DeleteObject_ReqParaStruct sReqDeleteFile;
    memset(&sReqDeleteFile,0,sizeof(Hce_WiFi_Command_Execute_DeleteObject_ReqParaStruct));
	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data, 0);

	ObjectID     = GetValueFromNameValueList(&data, "ObjectID");
    
    if(ObjectID == NULL)  {
        DLNA_ErrorPrintf("ERROR can't get Object ID\n");
		SoapError(h, 402, "Invalid Args");
        goto DestroyObject;
    }    
    else  {
       if((ExamFileExist(ObjectID, FileFullPath)) == 0)  {
           DPRINTF(E_DEBUG, L_HTTP, "Full Path: %s\n", FileFullPath);
           strncpyt(sReqDeleteFile.DataPathString,FileFullPath,HCE_WIFI_CAMERA_DATAPATH_STRLEN);
           HceDLNA_GeneralMessageToWiFi(HCE_WIFI_MSGID_CGI_PLAYCMD_DELETE_FILE,
                                        &sReqDeleteFile,
                                        sizeof(Hce_WiFi_Command_Execute_DeleteObject_ReqParaStruct),
                                        NULL,
                                        0);
        }
       else  {
          goto DestroyObject;
       }
    }
    
	str.off = sprintf(str.data, "%s", resp0);
	BuildSendAndCloseSoapResp(h, str.data, str.off);
DestroyObject:
	ClearNameValueList(&data);
    if(str.data != NULL)  {
        free(str.data);
    }    
    
}

#endif
    
static void
BrowseContentDirectory(struct upnphttp * h, const char * action)
{
	static const char resp0[] =
			"<u:BrowseResponse "
			"xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\">"
			"<Result>"
			"&lt;DIDL-Lite"
			CONTENT_DIRECTORY_SCHEMAS;
	struct magic_container_s *magic;
	char *zErrMsg = NULL;
	char *sql, *ptr;
	struct Response args;
	struct string_s str;
	int totalMatches = 0;
	int ret;
	const char *ObjectID, *BrowseFlag;
	char *Filter, *SortCriteria;
	const char *objectid_sql = "o.OBJECT_ID";
	const char *parentid_sql = "o.PARENT_ID";
	const char *refid_sql = "o.REF_ID";
	char where[256] = "";
	char *orderBy = NULL;
	struct NameValueParserData data;
	int RequestedCount = 0;
	int StartingIndex = 0;

	memset(&args, 0, sizeof(args));
	memset(&str, 0, sizeof(str));

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data, 0);

	ObjectID     = GetValueFromNameValueList(&data, "ObjectID");
	Filter       = GetValueFromNameValueList(&data, "Filter");
	BrowseFlag   = GetValueFromNameValueList(&data, "BrowseFlag");
	SortCriteria = GetValueFromNameValueList(&data, "SortCriteria");

	if( (ptr = GetValueFromNameValueList(&data, "RequestedCount")) )
		RequestedCount = atoi(ptr);
	if( RequestedCount < 0 )
	{
		SoapError(h, 402, "Invalid Args");
		goto browse_error;
	}
	if( !RequestedCount )
		RequestedCount = -1;
	if( (ptr = GetValueFromNameValueList(&data, "StartingIndex")) )
		StartingIndex = atoi(ptr);
	if( StartingIndex < 0 )
	{
		SoapError(h, 402, "Invalid Args");
		goto browse_error;
	}
	if( !BrowseFlag || (strcmp(BrowseFlag, "BrowseDirectChildren") && strcmp(BrowseFlag, "BrowseMetadata")) )
	{
		SoapError(h, 402, "Invalid Args");
		goto browse_error;
	}
	if( !ObjectID && !(ObjectID = GetValueFromNameValueList(&data, "ContainerID")) )
	{
		SoapError(h, 402, "Invalid Args");
		goto browse_error;
	}

	str.data = malloc(DEFAULT_RESP_SIZE);
	str.size = DEFAULT_RESP_SIZE;
    //----------------------------------------------------------------------------------------------
    //-------------------------------Construct BrowseResponse---------------------------------------
    //----------------------------------------------------------------------------------------------
	str.off = sprintf(str.data, "%s", resp0);
	/* See if we need to include DLNA namespace reference */
	args.iface = h->iface;
    //----------------------------------------------------------------------------------------------
    //-------------------------------Retrive filer and class---------------------------------------
    //----------------------------------------------------------------------------------------------
	args.filter = set_filter_flags(Filter, h);
	if( args.filter & FILTER_DLNA_NAMESPACE )
		ret = strcatf(&str, DLNA_NAMESPACE);
	if( args.filter & FILTER_PV_SUBTITLE )
		ret = strcatf(&str, PV_NAMESPACE);	
	strcatf(&str, "&gt;\n");

	args.returned = 0;
	args.requested = RequestedCount;
	args.client = h->req_client ? h->req_client->type->type : 0;
	args.flags = h->req_client ? h->req_client->type->flags : 0;
	args.str = &str;
	DPRINTF(E_DEBUG, L_HTTP, "Browsing ContentDirectory:\n"
	                         " * ObjectID: %s\n"
	                         " * Count: %d\n"
	                         " * StartingIndex: %d\n"
	                         " * BrowseFlag: %s\n"
	                         " * Filter: %s\n"
	                         " * SortCriteria: %s\n",
				ObjectID, RequestedCount, StartingIndex,
	                        BrowseFlag, Filter, SortCriteria);
//---------------------------------------------------------------------------------------------------
//----------------------------------------Metadata---------------------------------------------------
//---------------------------------------------------------------------------------------------------
	if( strcmp(BrowseFlag+6, "Metadata") == 0 )
	{

      #ifdef ABILITY_MODIFICATION
        if(bReplaceBrowseResponse == 1)  {
          UpnpSoapBrowseContainerStructure sInputInfo;
          memset(&sInputInfo,0,sizeof(UpnpSoapBrowseContainerStructure));
          strcpy(sInputInfo.pbID,"0");
          strcpy(sInputInfo.pbParent,"-1");
          strcpy(sInputInfo.pbTitle,"root");
          strcpy(sInputInfo.pbClass,"container");
          sInputInfo.eContainerType = DLNA_WIFI_BROWSE_CONTAINER_TYPE_METADATA;
          ret = UpnpSoapConstructContainerData((void *) &args,&sInputInfo);
          args.returned = 1;
    	  totalMatches  = args.returned;
        }
        else  {
        #ifndef UNUSE_SQL    
          //---------------------------------------------------------------------------------------------------
          //3------------------------------------Replace it--------------------------------------------
          //---------------------------------------------------------------------------------------------------
    		const char *id = ObjectID;
    		args.requested = 1;
    		magic = in_magic_container(ObjectID, args.flags, &id);
    		if (magic)
    		{
    			if (magic->objectid_sql && strcmp(id, ObjectID) != 0)
    				objectid_sql = magic->objectid_sql;
    			if (magic->parentid_sql && strcmp(id, ObjectID) != 0)
    				parentid_sql = magic->parentid_sql;
    			if (magic->refid_sql)
    				refid_sql = magic->refid_sql;
    		}
          //---------------------------------------------------------------------------------------------------
          //3------------------------------------Replace it--------------------------------------------
          //---------------------------------------------------------------------------------------------------
    		sql = sqlite3_mprintf("SELECT %s, %s, %s, " COLUMNS
    				      "from OBJECTS o left join DETAILS d on (d.ID = o.DETAIL_ID)"
    				      " where OBJECT_ID = '%q';",
    				      objectid_sql, parentid_sql, refid_sql, id);
    		ret = sqlite3_exec(db, sql, UpnpSoapCallback, (void *) &args, &zErrMsg);
          //---------------------------------------------------------------------------------------------------
          //----------------------------------------browse child---------------------------------------------------
          //---------------------------------------------------------------------------------------------------
    		totalMatches = args.returned;
        #endif  // ifndef UNUSE_SQL
        }
      #endif
	}
//---------------------------------------------------------------------------------------------------
//----------------------------------------browse child---------------------------------------------------
//---------------------------------------------------------------------------------------------------
	else
	{
		magic = check_magic_container(ObjectID, args.flags);
		if (magic)
		{
			if (magic->objectid && *(magic->objectid))
				ObjectID = *(magic->objectid);
			if (magic->objectid_sql)
				objectid_sql = magic->objectid_sql;
			if (magic->parentid_sql)
				parentid_sql = magic->parentid_sql;
			if (magic->refid_sql)
				refid_sql = magic->refid_sql;
			if (magic->where)
				strncpyt(where, magic->where, sizeof(where));
			if (magic->orderby && !GETFLAG(DLNA_STRICT_MASK))
				orderBy = strdup(magic->orderby);
			if (magic->max_count > 0)
			{
				int limit = MAX(magic->max_count - StartingIndex, 0);
				ret = get_child_count(ObjectID, magic);
				totalMatches = MIN(ret, limit);
				if (RequestedCount > limit || RequestedCount < 0)
					RequestedCount = limit;
			}
		}
        #ifndef UNUSE_SQL
		if (!where[0])
			sqlite3_snprintf(sizeof(where), where, "PARENT_ID = '%q'", ObjectID);
        #endif
	   //---------------------------------------------------------
	   //-----------Get total matched number----------
	   //---------------------------------------------------------
		if (!totalMatches)
			totalMatches = get_child_count(ObjectID, magic);
		ret = 0;
		if (SortCriteria && !orderBy)
		{
			__SORT_LIMIT
			orderBy = parse_sort_criteria(SortCriteria, &ret);
		}
		else if (!orderBy)
		{
			if( strncmp(ObjectID, MUSIC_PLIST_ID, strlen(MUSIC_PLIST_ID)) == 0 )
			{
				if( strcmp(ObjectID, MUSIC_PLIST_ID) == 0 )
					ret = xasprintf(&orderBy, "order by d.TITLE");
				else
					ret = xasprintf(&orderBy, "order by length(OBJECT_ID), OBJECT_ID");
			}
			else if( args.flags & FLAG_FORCE_SORT )
			{
				__SORT_LIMIT
				ret = xasprintf(&orderBy, "order by o.CLASS, d.DISC, d.TRACK, d.TITLE");
			}
			/* LG TV ordering bug */
			else if( args.client == ELGDevice )
				ret = xasprintf(&orderBy, "order by o.CLASS, d.TITLE");
			else
				orderBy = parse_sort_criteria(SortCriteria, &ret);
			if( ret == -1 )
			{
				free(orderBy);
				orderBy = NULL;
				ret = 0;
			}
		}
		/* If it's a DLNA client, return an error for bad sort criteria */
		if( ret < 0 && ((args.flags & FLAG_DLNA) || GETFLAG(DLNA_STRICT_MASK)) )
		{
			SoapError(h, 709, "Unsupported or invalid sort criteria");
			goto browse_error;
		}

      //---------------------------------------------------------------------------------------------------
      //3------------------------------------Replace it--------------------------------------------
      //---------------------------------------------------------------------------------------------------
      #ifdef ABILITY_MODIFICATION
        //4. Send Message to retrieve file information
        //Construct XML response
        if(bReplaceBrowseResponse == 1)  {
          UpnpBrowseResponseInputStruct  sInputInfo;
          UpnpBrowseResponseOutputStruct sOutputInfo;
          memset(&sInputInfo,0,sizeof(UpnpBrowseResponseInputStruct));
          memset(&sOutputInfo,0,sizeof(UpnpBrowseResponseOutputStruct));
          strcpy(sInputInfo.pbObjectID,ObjectID);
          if(strcmp(ObjectID, "0") == 0)  {
              ret = UpnpSoapBrowseResponse((void *) &args,&sInputInfo,&sOutputInfo);
          }
          else  {
              ret = UpnpSoapContainerBrowseResponse((void *) &args,&sInputInfo,&sOutputInfo, StartingIndex);
          }
          args.returned = sOutputInfo.ulCurrentReturnNum;
          totalMatches  = sOutputInfo.ulTotalMatchNum;
        }
        else  {
        #ifndef UNUSE_SQL    
          sql = sqlite3_mprintf("SELECT %s, %s, %s, " COLUMNS
                 "from OBJECTS o left join DETAILS d on (d.ID = o.DETAIL_ID)"
                 " where %s %s limit %d, %d;",
                 objectid_sql, parentid_sql, refid_sql,
                 where, THISORNUL(orderBy), StartingIndex, RequestedCount);
          DPRINTF(E_DEBUG, L_HTTP, "Browse SQL: %s\n", sql);
          ret = sqlite3_exec(db, sql, UpnpSoapCallback, (void *) &args, &zErrMsg);
        #endif  
        }
      #else
		sql = sqlite3_mprintf("SELECT %s, %s, %s, " COLUMNS
		                      "from OBJECTS o left join DETAILS d on (d.ID = o.DETAIL_ID)"
				      " where %s %s limit %d, %d;",
				      objectid_sql, parentid_sql, refid_sql,
				      where, THISORNUL(orderBy), StartingIndex, RequestedCount);
		DPRINTF(E_DEBUG, L_HTTP, "Browse SQL: %s\n", sql);
		ret = sqlite3_exec(db, sql, UpnpSoapCallback, (void *) &args, &zErrMsg);
  
      #endif
      //---------------------------------------------------------------------------------------------------
      //---------------------------------------------------------------------------------------------------
      //---------------------------------------------------------------------------------------------------
	}
#ifdef ABILITY_MODIFICATION    
    if(bReplaceBrowseResponse == 1)  {
        // Do nonthing
    }
    else  {
        #ifndef UNUSE_SQL
    	if( (ret != SQLITE_OK) && (zErrMsg != NULL) )
    	{
    		DPRINTF(E_WARN, L_HTTP, "SQL error: %s\nBAD SQL: %s\n", zErrMsg, sql);
    		sqlite3_free(zErrMsg);
    		SoapError(h, 709, "Unsupported or invalid sort criteria");
    		goto browse_error;
    	}
        DLNA_Printf(E_DEBUG, L_GENERAL,"[PPD] DBG\n");
    	sqlite3_free(sql);
        #endif
    }    
#else
	if( (ret != SQLITE_OK) && (zErrMsg != NULL) )
	{
		DPRINTF(E_WARN, L_HTTP, "SQL error: %s\nBAD SQL: %s\n", zErrMsg, sql);
		sqlite3_free(zErrMsg);
		SoapError(h, 709, "Unsupported or invalid sort criteria");
		goto browse_error;
	}
	sqlite3_free(sql);
#endif  
	/* Does the object even exist? */
	if( !totalMatches )
	{
		if( !object_exists(ObjectID) )
		{
			SoapError(h, 701, "No such object error");
			goto browse_error;
		}
	}
	ret = strcatf(&str, "&lt;/DIDL-Lite&gt;</Result>\n"
	                    "<NumberReturned>%u</NumberReturned>\n"
	                    "<TotalMatches>%u</TotalMatches>\n"
	                    "<UpdateID>%u</UpdateID>"
	                    "</u:BrowseResponse>",
	                    args.returned, totalMatches, updateID);
	BuildSendAndCloseSoapResp(h, str.data, str.off);
browse_error:
	ClearNameValueList(&data);
	free(orderBy);
	free(str.data);
}

static inline void
charcat(struct string_s *str, char c)
{
	if (str->size <= str->off)
	{
		str->data[str->size-1] = '\0';
		return;
	}
	str->data[str->off] = c;
	str->off += 1;
}

static inline char *
parse_search_criteria(const char *str, char *sep)
{
	struct string_s criteria;
	int len;
	int literal = 0, like = 0;
	const char *s;

	if (!str)
		return strdup("1 = 1");

	len = strlen(str) + 32;
	criteria.data = malloc(len);
	criteria.size = len;
	criteria.off = 0;

	s = str;

	while (isspace(*s))
		s++;

	while (*s)
	{
		if (literal)
		{
			switch (*s) {
			case '&':
				if (strncmp(s, "&quot;", 6) == 0)
					s += 5;
				else if (strncmp(s, "&apos;", 6) == 0)
				{
					strcatf(&criteria, "'");
					s += 6;
					continue;
				}
				else
					break;
			case '"':
				literal = 0;
				if (like)
				{
					charcat(&criteria, '%');
					like--;
				}
				charcat(&criteria, '"');
				break;
			case '\\':
				if (strncmp(s, "\\&quot;", 7) == 0)
				{
					strcatf(&criteria, "&amp;quot;");
					s += 7;
					continue;
				}
				break;
			case 'o':
				if (strncmp(s, "object.", 7) == 0)
					s += 7;
				else if (strncmp(s, "object\"", 7) == 0 ||
				         strncmp(s, "object&quot;", 12) == 0)
				{
					s += 6;
					continue;
				}
			default:
				charcat(&criteria, *s);
				break;
			}
		}
		else
		{
			switch (*s) {
			case '\\':
				if (strncmp(s, "\\&quot;", 7) == 0)
				{
					strcatf(&criteria, "&amp;quot;");
					s += 7;
					continue;
				}
				else
					charcat(&criteria, *s);
				break;
			case '"':
				literal = 1;
				charcat(&criteria, *s);
				if (like == 2)
				{
					charcat(&criteria, '%');
					like--;
				}
				break;
			case '&':
				if (strncmp(s, "&quot;", 6) == 0)
				{
					literal = 1;
					strcatf(&criteria, "\"");
					if (like == 2)
					{
						charcat(&criteria, '%');
						like--;
					}
					s += 5;
				}
				else if (strncmp(s, "&apos;", 6) == 0)
				{
					strcatf(&criteria, "'");
					s += 5;
				}
				else if (strncmp(s, "&lt;", 4) == 0)
				{
					strcatf(&criteria, "<");
					s += 3;
				}
				else if (strncmp(s, "&gt;", 4) == 0)
				{
					strcatf(&criteria, ">");
					s += 3;
				}
				else
					charcat(&criteria, *s);
				break;
			case '@':
				if (strncmp(s, "@refID", 6) == 0)
				{
					strcatf(&criteria, "REF_ID");
					s += 6;
					continue;
				}
				else if (strncmp(s, "@id", 3) == 0)
				{
					strcatf(&criteria, "OBJECT_ID");
					s += 3;
					continue;
				}
				else if (strncmp(s, "@parentID", 9) == 0)
				{
					strcatf(&criteria, "PARENT_ID");
					s += 9;
					strcpy(sep, "*");
					continue;
				}
				else
					charcat(&criteria, *s);
				break;
			case 'c':
				if (strncmp(s, "contains", 8) == 0)
				{
					strcatf(&criteria, "like");
					s += 8;
					like = 2;
					continue;
				}
				else
					charcat(&criteria, *s);
				break;
			case 'd':
				if (strncmp(s, "derivedfrom", 11) == 0)
				{
					strcatf(&criteria, "like");
					s += 11;
					like = 1;
					continue;
				}
				else if (strncmp(s, "dc:date", 7) == 0)
				{
					strcatf(&criteria, "d.DATE");
					s += 7;
					continue;
				}
				else if (strncmp(s, "dc:title", 8) == 0)
				{
					strcatf(&criteria, "d.TITLE");
					s += 8;
					continue;
				}
				else if (strncmp(s, "dc:creator", 10) == 0)
				{
					strcatf(&criteria, "d.CREATOR");
					s += 10;
					continue;
				}
				else
					charcat(&criteria, *s);
				break;
			case 'e':
				if (strncmp(s, "exists", 6) == 0)
				{
					s += 6;
					while (isspace(*s))
						s++;
					if (strncmp(s, "true", 4) == 0)
					{
						strcatf(&criteria, "is not NULL");
						s += 3;
					}
					else if (strncmp(s, "false", 5) == 0)
					{
						strcatf(&criteria, "is NULL");
						s += 4;
					}
				}
				else
					charcat(&criteria, *s);
				break;
			case 'u':
				if (strncmp(s, "upnp:class", 10) == 0)
				{
					strcatf(&criteria, "o.CLASS");
					s += 10;
					continue;
				}
				else if (strncmp(s, "upnp:actor", 10) == 0)
				{
					strcatf(&criteria, "d.ARTIST");
					s += 10;
					continue;
				}
				else if (strncmp(s, "upnp:artist", 11) == 0)
				{
					strcatf(&criteria, "d.ARTIST");
					s += 11;
					continue;
				}
				else if (strncmp(s, "upnp:album", 10) == 0)
				{
					strcatf(&criteria, "d.ALBUM");
					s += 10;
					continue;
				}
				else if (strncmp(s, "upnp:genre", 10) == 0)
				{
					strcatf(&criteria, "d.GENRE");
					s += 10;
					continue;
				}
				else
					charcat(&criteria, *s);
				break;
			case '(':
				if (s > str && !isspace(s[-1]))
					charcat(&criteria, ' ');
				charcat(&criteria, *s);
				break;
			case ')':
				charcat(&criteria, *s);
				if (!isspace(s[1]))
					charcat(&criteria, ' ');
				break;
			default:
				charcat(&criteria, *s);
				break;
			}
		}
		s++;
	}
	charcat(&criteria, '\0');

	return criteria.data;
}

static void
SearchContentDirectory(struct upnphttp * h, const char * action)
{
// remove it
#ifndef ABILITY_MODIFICATION
	static const char resp0[] =
			"<u:SearchResponse "
			"xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\">"
			"<Result>"
			"&lt;DIDL-Lite"
			CONTENT_DIRECTORY_SCHEMAS;
	struct magic_container_s *magic;
	char *zErrMsg = NULL;
	char *sql, *ptr;
	struct Response args;
	struct string_s str;
	int totalMatches;
	int ret;
	const char *ContainerID;
	char *Filter, *SearchCriteria, *SortCriteria;
	char *orderBy = NULL, *where = NULL, sep[] = "$*";
	char groupBy[] = "group by DETAIL_ID";
	struct NameValueParserData data;
	int RequestedCount = 0;
	int StartingIndex = 0;

	memset(&args, 0, sizeof(args));
	memset(&str, 0, sizeof(str));

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data, 0);

	ContainerID = GetValueFromNameValueList(&data, "ContainerID");
	Filter = GetValueFromNameValueList(&data, "Filter");
	SearchCriteria = GetValueFromNameValueList(&data, "SearchCriteria");
	SortCriteria = GetValueFromNameValueList(&data, "SortCriteria");

	if( (ptr = GetValueFromNameValueList(&data, "RequestedCount")) )
		RequestedCount = atoi(ptr);
	if( !RequestedCount )
		RequestedCount = -1;
	if( (ptr = GetValueFromNameValueList(&data, "StartingIndex")) )
		StartingIndex = atoi(ptr);
	if( !ContainerID )
	{
		if( !(ContainerID = GetValueFromNameValueList(&data, "ObjectID")) )
		{
			SoapError(h, 402, "Invalid Args");
			goto search_error;
		}
	}

	str.data = malloc(DEFAULT_RESP_SIZE);
	str.size = DEFAULT_RESP_SIZE;
	str.off = sprintf(str.data, "%s", resp0);
	/* See if we need to include DLNA namespace reference */
	args.iface = h->iface;
	args.filter = set_filter_flags(Filter, h);
	if( args.filter & FILTER_DLNA_NAMESPACE )
	{
		ret = strcatf(&str, DLNA_NAMESPACE);
	}
	strcatf(&str, "&gt;\n");

	args.returned = 0;
	args.requested = RequestedCount;
	args.client = h->req_client ? h->req_client->type->type : 0;
	args.flags = h->req_client ? h->req_client->type->flags : 0;
	args.str = &str;
	DPRINTF(E_DEBUG, L_HTTP, "Searching ContentDirectory:\n"
	                         " * ObjectID: %s\n"
	                         " * Count: %d\n"
	                         " * StartingIndex: %d\n"
	                         " * SearchCriteria: %s\n"
	                         " * Filter: %s\n"
	                         " * SortCriteria: %s\n",
				ContainerID, RequestedCount, StartingIndex,
	                        SearchCriteria, Filter, SortCriteria);

	magic = check_magic_container(ContainerID, args.flags);
	if (magic && magic->objectid && *(magic->objectid))
		ContainerID = *(magic->objectid);

	if( strcmp(ContainerID, "0") == 0 )
		ContainerID = "*";

	if( strcmp(ContainerID, MUSIC_ALL_ID) == 0 ||
	    GETFLAG(DLNA_STRICT_MASK) )
		groupBy[0] = '\0';

	where = parse_search_criteria(SearchCriteria, sep);
	DPRINTF(E_DEBUG, L_HTTP, "Translated SearchCriteria: %s\n", where);

	totalMatches = sql_get_int_field(db, "SELECT (select count(distinct DETAIL_ID)"
	                                     " from OBJECTS o left join DETAILS d on (o.DETAIL_ID = d.ID)"
	                                     " where (OBJECT_ID glob '%q%s') and (%s))"
	                                     " + "
	                                     "(select count(*) from OBJECTS o left join DETAILS d on (o.DETAIL_ID = d.ID)"
	                                     " where (OBJECT_ID = '%q') and (%s))",
	                                     ContainerID, sep, where, ContainerID, where);
	if( totalMatches < 0 )
	{
		/* Must be invalid SQL, so most likely bad or unhandled search criteria. */
		SoapError(h, 708, "Unsupported or invalid search criteria");
		goto search_error;
	}
	/* Does the object even exist? */
	if( !totalMatches )
	{
		if( !object_exists(ContainerID) )
		{
			SoapError(h, 710, "No such container");
			goto search_error;
		}
	}
	ret = 0;
	__SORT_LIMIT
	orderBy = parse_sort_criteria(SortCriteria, &ret);
	/* If it's a DLNA client, return an error for bad sort criteria */
	if( ret < 0 && ((args.flags & FLAG_DLNA) || GETFLAG(DLNA_STRICT_MASK)) )
	{
		SoapError(h, 709, "Unsupported or invalid sort criteria");
		goto search_error;
	}

	sql = sqlite3_mprintf( SELECT_COLUMNS
	                      "from OBJECTS o left join DETAILS d on (d.ID = o.DETAIL_ID)"
	                      " where OBJECT_ID glob '%q%s' and (%s) %s "
	                      "%z %s"
	                      " limit %d, %d",
	                      ContainerID, sep, where, groupBy,
	                      (*ContainerID == '*') ? NULL :
	                      sqlite3_mprintf("UNION ALL " SELECT_COLUMNS
	                                      "from OBJECTS o left join DETAILS d on (d.ID = o.DETAIL_ID)"
	                                      " where OBJECT_ID = '%q' and (%s) ", ContainerID, where),
	                      orderBy, StartingIndex, RequestedCount);
	DPRINTF(E_DEBUG, L_HTTP, "Search SQL: %s\n", sql);
	ret = sqlite3_exec(db, sql, UpnpSoapCallback, (void *) &args, &zErrMsg);
	if( (ret != SQLITE_OK) && (zErrMsg != NULL) )
	{
		DPRINTF(E_WARN, L_HTTP, "SQL error: %s\nBAD SQL: %s\n", zErrMsg, sql);
		sqlite3_free(zErrMsg);
	}
	sqlite3_free(sql);
	ret = strcatf(&str, "&lt;/DIDL-Lite&gt;</Result>\n"
	                    "<NumberReturned>%u</NumberReturned>\n"
	                    "<TotalMatches>%u</TotalMatches>\n"
	                    "<UpdateID>%u</UpdateID>"
	                    "</u:SearchResponse>",
	                    args.returned, totalMatches, updateID);
	BuildSendAndCloseSoapResp(h, str.data, str.off);
search_error:
	ClearNameValueList(&data);
	free(orderBy);
	free(where);
	free(str.data);

#endif
    
}

/*
If a control point calls QueryStateVariable on a state variable that is not
buffered in memory within (or otherwise available from) the service,
the service must return a SOAP fault with an errorCode of 404 Invalid Var.

QueryStateVariable remains useful as a limited test tool but may not be
part of some future versions of UPnP.
*/
static void
QueryStateVariable(struct upnphttp * h, const char * action)
{
	static const char resp[] =
        "<u:%sResponse "
        "xmlns:u=\"%s\">"
		"<return>%s</return>"
        "</u:%sResponse>";

	char body[512];
	struct NameValueParserData data;
	const char * var_name;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data, 0);
	/*var_name = GetValueFromNameValueList(&data, "QueryStateVariable"); */
	/*var_name = GetValueFromNameValueListIgnoreNS(&data, "varName");*/
	var_name = GetValueFromNameValueList(&data, "varName");

	DPRINTF(E_INFO, L_HTTP, "QueryStateVariable(%.40s)\n", var_name);

	if(!var_name)
	{
		SoapError(h, 402, "Invalid Args");
	}
	else if(strcmp(var_name, "ConnectionStatus") == 0)
	{	
		int bodylen;
		bodylen = snprintf(body, sizeof(body), resp,
                           action, "urn:schemas-upnp-org:control-1-0",
		                   "Connected", action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}
	else
	{
		DPRINTF(E_WARN, L_HTTP, "%s: Unknown: %s\n", action, THISORNUL(var_name));
		SoapError(h, 404, "Invalid Var");
	}

	ClearNameValueList(&data);	
}

static void
SamsungGetFeatureList(struct upnphttp * h, const char * action)
{
	static const char resp[] =
		"<u:X_GetFeatureListResponse xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\">"
		"<FeatureList>"
		"&lt;Features xmlns=\"urn:schemas-upnp-org:av:avs\""
		" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\""
		" xsi:schemaLocation=\"urn:schemas-upnp-org:av:avs http://www.upnp.org/schemas/av/avs.xsd\"&gt;"
		"&lt;Feature name=\"samsung.com_BASICVIEW\" version=\"1\"&gt;"
		 "&lt;container id=\"%s\" type=\"object.item.audioItem\"/&gt;"
		 "&lt;container id=\"%s\" type=\"object.item.videoItem\"/&gt;"
		 "&lt;container id=\"%s\" type=\"object.item.imageItem\"/&gt;"
		"&lt;/Feature&gt;"
		"&lt;/Features&gt;"
		"</FeatureList></u:X_GetFeatureListResponse>";
	const char *audio = MUSIC_ID;
	const char *video = VIDEO_ID;
	const char *image = IMAGE_ID;
	char body[1024];
	int len;

	if (runtime_vars.root_container)
	{
		if (strcmp(runtime_vars.root_container, BROWSEDIR_ID) == 0)
		{
			audio = MUSIC_DIR_ID;
			video = VIDEO_DIR_ID;
			image = IMAGE_DIR_ID;
		}
		else
		{
			audio = runtime_vars.root_container;
			video = runtime_vars.root_container;
			image = runtime_vars.root_container;
		}
	}

	len = snprintf(body, sizeof(body), resp, audio, video, image);

	BuildSendAndCloseSoapResp(h, body, len);
}

static void
SamsungSetBookmark(struct upnphttp * h, const char * action)
{
	static const char resp[] =
	    "<u:X_SetBookmarkResponse"
	    " xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\">"
	    "</u:X_SetBookmarkResponse>";

	struct NameValueParserData data;
	char *ObjectID, *PosSecond;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data, 0);
	ObjectID = GetValueFromNameValueList(&data, "ObjectID");
	PosSecond = GetValueFromNameValueList(&data, "PosSecond");

	if ( atoi(PosSecond) < 30 )
		PosSecond = "0";

	if( ObjectID && PosSecond )
	{
		int ret;
		const char *rid = ObjectID;

		in_magic_container(ObjectID, 0, &rid);
		ret = sql_exec(db, "INSERT OR REPLACE into BOOKMARKS"
		                   " VALUES "
		                   "((select DETAIL_ID from OBJECTS where OBJECT_ID = '%q'), %q)", rid, PosSecond);
		if( ret != SQLITE_OK )
			DPRINTF(E_WARN, L_METADATA, "Error setting bookmark %s on ObjectID='%s'\n", PosSecond, rid);
		BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
	}
	else
		SoapError(h, 402, "Invalid Args");

	ClearNameValueList(&data);	
}

#ifdef ABILITY_MODIFICATION

static void
BuildSendSoapReqest(struct upnphttp * h, char* action, enum service_types service_type,
                          const char * body,  int bodylen)
{
	static const char beforebody[] =
		"<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
		"<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" "
		"s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
		"<s:Body>\r\n";

	static const char afterbody[] =
		"</s:Body>\r\n"
		"</s:Envelope>\r\n";

	if (!body || bodylen < 0)
	{
		Send500(h);
        if(h->errHandlerEx){
            ErrHandler(h);
        }
		return;
	}

	BuildHeader_upnphttp_SOAP(h, action, service_type,  sizeof(beforebody) - 1
		+ sizeof(afterbody) - 1 + bodylen );

	memcpy(h->res_buf + h->res_buflen, beforebody, sizeof(beforebody) - 1);
	h->res_buflen += sizeof(beforebody) - 1;

	memcpy(h->res_buf + h->res_buflen, body, bodylen);
	h->res_buflen += bodylen;

	memcpy(h->res_buf + h->res_buflen, afterbody, sizeof(afterbody) - 1);
	h->res_buflen += sizeof(afterbody) - 1;

	SendRequest_SOAP(h);
}


void ErrHandlerEx_CreateObject(char* func, int line){
    DPRINTF(E_ERROR, L_HTTP, "ErrHandlerEx_CreateObject: %s[%d] error\n", func, line);
    if( Get_Upload_Status() != UPLOAD_STATUS_CANCELING)
        Notify_Transfer_Result(1);
    Set_Upload_Status(UPLOAD_STATUS_IDLE); 
}

void ErrHandlerEx_X_GetDLNAUploadProfile(char* func, int line){
    DPRINTF(E_DEBUG, L_HTTP, " %s[%d] error\n", func, line);
    /* do not thing*/
}


static void
upload_X_GetDLNAUploadProfiles(struct upnphttp * h, const char * action){

    static const char body[] =
        "    <u:X_GetDLNAUploadProfiles xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\">\r\n"
        "      <UploadProfiles>JPEG_SM,JPEG_MED,JPEG_LRG,PANASONIC.COM_PN=MPO,AVC_MP4_BL_L31_HD_AAC,AVC_MP4_MP_HD_1080i_AAC</UploadProfiles>\r\n"
        "    </u:X_GetDLNAUploadProfiles>\r\n";

    DPRINTF(E_DEBUG, L_HTTP,
"strlen: %d, sizeof(body): %d\n",strlen(body),sizeof(body) );

    h->req_soapSent = ESSOAP_X_GetDLNAUploadProfiles;
    
    BuildSendSoapReqest(h, action, EsContentDirectory,
                          body,  sizeof(body)-1 );
        
}

static void
upload_CreateObject(struct upnphttp * h, const char * action){

/* Only for testing now, need to follow the real case*/

    DPRINTF(E_WARN, L_HTTP,"upload_CreateObject\n");

    char item_name[20]={0};

    static const char req[] =
        "    <u:CreateObject xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\">\r\n"
        "     <ContainerID>DLNA.ORG_AnyContainer</ContainerID>\r\n"
        "      <Elements>"
        "&lt;DIDL-Lite xmlns=&quot;urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/&quot;"
        " xmlns:dc=&quot;http://purl.org/dc/elements/1.1/&quot;"
        " xmlns:upnp=&quot;urn:schemas-upnp-org:metadata-1-0/upnp/&quot;"
        " xmlns:pana=&quot;urn:schemas-panasonic-com:pana&quot;"
        " &gt;"
        " &lt;item id=&quot;&quot; parentID=&quot;DLNA.ORG_AnyContainer&quot; restricted=&quot;0&quot;"
        " &gt;"
        "&lt;dc:title&gt;"
        "%s" /* insert item name ,ex: P10600410002 */
        "&lt;/dc:title&gt;"
        "&lt;upnp:class&gt;"
        "object.item.%s" /* insert item type, ex: imageItem */
        "&lt;/upnp:class&gt;"
        "&lt;res protocolInfo="
        "&quot;*:*:%s&quot;" /* insert item info, ex: image/jpeg:DLNA.ORG_PN=JPEG_LRG;DLNA.ORG_CI=0 */
        " size=&quot;%d&quot;" /* insert data size, ex: 3621408, unit: byte */
        " resolution=&quot;%dx%d&quot;&gt;" /* insert resolution ,ex: 3456x2592 */
        "&lt;/res&gt;"
        "&lt;dc:date&gt;"
        "%s" /*insert create date,ex: 2017-05-13T16:28:49 */
        "&lt;/dc:date&gt;"
        /*"&lt;pana:X_Type&gt;"*/
        /*"&lt;/pana:X_Type&gt;"*/
        "&lt;/item&gt;"
        "&lt;/DIDL-Lite&gt;</Elements>\r\n"
        "    </u:CreateObject>"
        ;

    char body[2048]; 
    int bodylen;
    int meida_type;
    //Parse filePath
    char *p, *pn; 
    char protocolinfo[60] ={0};
    char filetype[24] = {0};
    char filename[10]={0};

    h->errHandlerEx = ErrHandlerEx_CreateObject;
    
    // get media type
    p = strchr (upload_object_info.content_full_path, '.');
    if(p == NULL){
        //Bad file path
        DPRINTF(E_ERROR, L_HTTP, "Bad file path: %s\n", upload_object_info.content_full_path);
        ErrHandler(h);
        h->state = 100;
        return;
    }
    
    //set media type
    if(strcmp("JPG", p+1) == 0 ){
        meida_type = TYPE_IMAGES;
    }    
    else if(strcmp("MP4", p+1) == 0 ){
        meida_type = TYPE_VIDEO;
    }
    else if(strcmp("MOV", p+1) == 0 ){
        meida_type = TYPE_VIDEO;
    }
    else{
        DPRINTF(E_ERROR, L_HTTP, "unknown media type\n");
        ErrHandler(h);
        h->state = 100;
        return;
    }

    //get filename
    pn = strrchr (upload_object_info.content_full_path, '/');
    if(pn == NULL){
        //Bad file path
        DPRINTF(E_ERROR, L_HTTP, "Bad file path: %s\n", upload_object_info.content_full_path);
        ErrHandler(h);
        h->state = 100;
        return;
    }
    
    strncpy(filename, pn+1, p-pn-1);
    DPRINTF(E_DEBUG, L_HTTP, "itemname %s\n",filename);

    //set protocol info
    if(meida_type == TYPE_IMAGES){
        //"image/jpeg:DLNA.ORG_PN=JPEG_LRG;DLNA.ORG_CI=0"
        if(upload_object_info.nWidth > 4096 || upload_object_info.nHeight > 4096)  {
              strcpy(filetype," ");
        }
        else if(upload_object_info.nWidth >= 1920 || upload_object_info.nHeight >= 1080)  {
            strcpy(filetype,"DLNA.ORG_PN=JPEG_LRG");
        }
        else if(upload_object_info.nWidth >= 1024 || upload_object_info.nHeight >= 768)  {
            strcpy(filetype,"DLNA.ORG_PN=JPEG_LRG");
        }
        else if(upload_object_info.nWidth >= 640 || upload_object_info.nHeight >= 480)  {
            strcpy(filetype,"DLNA.ORG_PN=JPEG_MED");
        }
        else {
            strcpy(filetype,"DLNA.ORG_PN=JPEG_SM");
        }
        sprintf(protocolinfo,"image/jpeg:%s,DLNA.ORG_CI=%d", filetype, upload_object_info.resize);
        
    }else if(meida_type == TYPE_VIDEO){
        //"application/octet-stream:DLNA.ORG_CI=0"
        sprintf(protocolinfo,"application/octet-stream:DLNA.ORG_CI=0");
    }
    
    bodylen = snprintf(body, sizeof(body), req,
		filename,
		(meida_type == TYPE_IMAGES)? "imageItem":"videoItem.movie",
		protocolinfo,
		upload_object_info.nBufferSize,
		upload_object_info.nWidth,
		upload_object_info.nHeight,
		"2017-05-13T16:28:49"
		);

    h->req_soapSent = ESSOAP_CreateObject;

    BuildSendSoapReqest(h, action, EsContentDirectory,
                          body,  bodylen );

}

int Process_Res_X_GetDLNAUploadProfiles(struct upnphttp *h, struct NameValueParserData* pxml){
    
    char* supported;

    supported = GetValueFromNameValueList(pxml, "SupportedUploadProfiles");

    DPRINTF(E_DEBUG, L_HTTP, "SupportedUploadProfiles: %s\n", supported);
    /* Todo : check the values, and record?*/

    h->state = 100; //done, close socket
    
    return 0;
}

int Process_Res_CreateObject(struct upnphttp *h, struct NameValueParserData* pxml){

    char* result;
    char *p, *t, *port, *urlend;

    result = GetValueFromNameValueList(pxml, "Result");

    if(result == NULL){
        DPRINTF(E_ERROR, L_HTTP, "Process_Res_CreateObject: unexpect response\n");
        if(h->errHandlerEx){
            ErrHandler(h);
        }
        h->state = 100;
        return -1;
    }
    // importUri=&quot;http://192.168.54.10:49153/MEDIA-027-i-1&quot;
    p = strstr( result, "importUri");
    if( p == NULL ){
        DPRINTF(E_DEBUG, L_HTTP, "Process_Res_CreateObject: miss importUri\n");
        if(h->errHandlerEx){
            ErrHandler(h);
        }
        h->state = 100;
        return -1;
    }

    p = strstr( p, "http");
    p += 7;


    port = strchr( p, ':');
 
    t = strchr( p, '/');
    if( t == NULL ){
        DPRINTF(E_ERROR, L_HTTP, "Process_Res_CreateObject: miss importUri\n");
        if(h->errHandlerEx){
            ErrHandler(h);
        }
        h->state = 100;
        return -1;
    }
    
    *t = '\0';
    t += 1;

    urlend = strstr( t, "&quot");  
    *urlend = '\0';

    h->POST_path = strdup(t);//FIXME: need free, please check.

    if( port == NULL ){
        h->POST_target_port = 80;
        DPRINTF(E_DEBUG, L_HTTP, "Process_Res_CreateObject: miss port, use port:80\n");
    }else{
        port = port+1;
        h->POST_target_port = atoi(port);
    }

    DPRINTF(E_DEBUG, L_HTTP, "Process_Res_CreateObject: path: %s, port: %ld\n", h->POST_path, h->POST_target_port);

    h->state = PREPARE_SEND_POST_DATA;

    return 0;
}



#endif


static const struct 
{
	const char * methodName; 
	void (*methodImpl)(struct upnphttp *, const char *);
}
soapMethods[] =
{
	{ "QueryStateVariable", QueryStateVariable},
	{ "Browse", BrowseContentDirectory},
	{ "Search", SearchContentDirectory},
	{ "GetSearchCapabilities", GetSearchCapabilities},
	{ "GetSortCapabilities", GetSortCapabilities},
	{ "GetSystemUpdateID", GetSystemUpdateID},
	{ "GetProtocolInfo", GetProtocolInfo},
	{ "GetCurrentConnectionIDs", GetCurrentConnectionIDs},
	{ "GetCurrentConnectionInfo", GetCurrentConnectionInfo},
	{ "IsAuthorized", IsAuthorizedValidated},
	{ "IsValidated", IsAuthorizedValidated},
	{ "RegisterDevice", RegisterDevice},
	{ "X_GetFeatureList", SamsungGetFeatureList},
	{ "X_SetBookmark", SamsungSetBookmark},
#ifdef ABILITY_MODIFICATION	
	{ "DestroyObject", ExecDestroyObject},
#endif	
	{ 0, 0 }
};

#ifdef ABILITY_MODIFICATION
static const struct 
{
	const char * methodName; 
	void (*methodImpl)(struct upnphttp *, const char *);
    int (*methodImpl_res)(struct upnphttp *, struct NameValueParserData*);
}
upload_soapMethods[] =
{
	{ "X_GetDLNAUploadProfiles", upload_X_GetDLNAUploadProfiles, Process_Res_X_GetDLNAUploadProfiles},
    { "CreateObject", upload_CreateObject, Process_Res_CreateObject},
	{ 0, 0 }
};
#endif


void
ExecuteSoapAction(struct upnphttp * h, const char * action, int n)
{
	char * p;

	p = strchr(action, '#');
	if(p)
	{
		int i = 0;
		int len;
		int methodlen;
		char * p2;
		p++;
		p2 = strchr(p, '"');
		if(p2)
			methodlen = p2 - p;
		else
			methodlen = n - (p - action);
		DPRINTF(E_DEBUG, L_HTTP, "SoapMethod: %.*s\n", methodlen, p);
		while(soapMethods[i].methodName)
		{
			len = strlen(soapMethods[i].methodName);
			if(strncmp(p, soapMethods[i].methodName, len) == 0)
			{
				soapMethods[i].methodImpl(h, soapMethods[i].methodName);
				return;
			}
			i++;
		}

		DPRINTF(E_WARN, L_HTTP, "SoapMethod: Unknown: %.*s\n", methodlen, p);
	}

	SoapError(h, 401, "Invalid Action");
}

#ifdef ABILITY_MODIFICATION

/* Prepare and return the xml body for SOAP */
void
Prepare_SoapAction(struct upnphttp * h, const char * action){

    int i=0;
    int len;
    
    while(upload_soapMethods[i].methodName)
    {
		len = strlen(upload_soapMethods[i].methodName);
		if(strncmp(action, upload_soapMethods[i].methodName, len) == 0)
		{
			upload_soapMethods[i].methodImpl(h, upload_soapMethods[i].methodName);
			return;
		}
		i++;
	} 
    
    DPRINTF(E_WARN, L_HTTP, "upload_soapMethods: Unknown: %s\n", action);
    
}

void Process_SoapAction_Res(struct upnphttp * h, struct NameValueParserData *pxml){

    upload_soapMethods[h->req_soapSent].methodImpl_res( h, pxml);
    
}

#endif


